# (c) 2015-2018 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#

from moleculekit.molecule import Molecule
from htmd.protocols.oldprotocolinterface import ProtocolInterface, TYPE_INT, TYPE_FLOAT, RANGE_0POS, RANGE_POS, RANGE_ANY
from htmd.apps.acemd import Acemd
import os
import htmd
import logging
from htmd.decorators import _Deprecated
logger = logging.getLogger(__name__)


@_Deprecated('1.13.6')
class Production(ProtocolInterface):
    ''' Production protocol v3

        Production protocol for globular and membrane proteins
        It also includes a possible flatbottom potential box
        It is also possible to define constraints for the production run

        Parameters
        ----------
        runtime : float, default=0
            Running time of the simulation.
        timeunits : str, default='steps'
            Units for runtime. Can be 'steps', 'ns' etc.
        temperature : float, default=300
            Temperature of the thermostat in Kelvin
        fb_k : float, default=0
            Force constant of the flatbottom potential in kcal/mol/A^2. E.g. 5
        fb_reference : str, default='none'
            Reference selection to use as dynamic center of the flatbottom box.
        fb_selection : str, default='none'
            Selection of atoms to apply the flatbottom potential
        fb_box : list, default=[0, 0, 0, 0, 0, 0]
            Position of the flatbottom box in term of the reference center given as [xmin, xmax, ymin, ymax, zmin, zmax]
        useconstraints: bool, default=False
            Apply constraints to the production simulation, defined by the constraints parameter
        constraints : dict, default={}
            A dictionary of atomselections and values of the constraint to be applied (in kcal/mol/A^2). Atomselects must be mutually exclusive.

    '''
    def __init__(self):
        super().__init__()
        self._cmdObject('acemd', ':class:`MDEngine <htmd.apps.app.App>` object', 'MD engine', None, Acemd)
        self._cmdValue('runtime', 'float', 'Running time of the simulation.', 0, TYPE_FLOAT, RANGE_0POS)
        self._cmdString('timeunits', 'str', 'Units for runtime. Can be \'steps\', \'ns\' etc.', 'steps')
        self._cmdValue('temperature', 'float', 'Temperature of the thermostat in Kelvin', 300, TYPE_FLOAT, RANGE_ANY)
        self._cmdValue('fb_k', 'float', 'Force constant of the flatbottom potential in kcal/mol/A^2. E.g. 5', 0, TYPE_FLOAT, RANGE_ANY)
        self._cmdString('fb_reference', 'str', 'Reference selection to use as dynamic center of the flatbottom box.', 'none')
        self._cmdString('fb_selection', 'str', 'Selection of atoms to apply the flatbottom potential', 'none')
        self._cmdList('fb_box', 'list', 'Position of the flatbottom box in term of the reference center given as [xmin, xmax, ymin, ymax, zmin, zmax]', [0,0,0,0,0,0])
        self._cmdList('useconstraints', 'bool', 'Apply constraints to the production simulation, defined by the constraints parameter', False)
        self._cmdDict('constraints', 'dict', 'A dictionary of atomselections and values of the constraint to be applied '
                                             '(in kcal/mol/A^2). Atomselects must be mutually exclusive.'
                                             , {})

        self.acemd = Acemd()
        #self.acemd.binindex='input.idx'
        self.acemd.extendedsystem='input.xsc'
        self.acemd.coordinates = None
        self.acemd.structure = None
        self.acemd.parameters = None
        self.acemd.temperature = '300'
        self.acemd.restart = 'on'
        self.acemd.restartfreq = '5000'
        self.acemd.outputname = 'output'
        self.acemd.xtcfile = 'output.xtc'
        self.acemd.xtcfreq = '25000'
        self.acemd.timestep = '4'
        self.acemd.rigidbonds = 'all'
        self.acemd.hydrogenscale = '4'
        self.acemd.switching = 'on'
        self.acemd.switchdist = '7.5'
        self.acemd.cutoff = '9'
        self.acemd.exclude = 'scaled1-4'
        self.acemd.scaling14 = '1.0'
        self.acemd.langevin = 'on'
        self.acemd.langevintemp = '300'
        self.acemd.langevindamping = '0.1'
        self.acemd.pme = 'on'
        self.acemd.pmegridspacing = '1.0'
        self.acemd.fullelectfrequency = '2'
        self.acemd.energyfreq = '5000'
        self.acemd.run = '$numsteps'
        self.acemd.TCL = ('''
set numsteps {NUMSTEPS}
set fb_refindex {{ {REFINDEX} }}
set fb_selindex {{ {SELINDEX} }}
set fb_box {{ {BOX} }}
set fb_K {KCONST}
#
''',
'''
proc flatbot1d {x xm xM fb_K} {
  set f 0
  if {$x < $xm} {
    set f [expr $fb_K*[expr $xm-$x]]
  }
  if {$x > $xM} {
    set f [expr $fb_K*[expr $xM-$x]]
  }
  return $f
}
proc calcforces_init {} {
  global ref sel fb_refindex fb_selindex
  berendsenpressure  off
  set ref [addgroup  $fb_refindex]
  set sel [addgroup  $fb_selindex]
}
proc calcforces {} {
  global ref sel fb_K fb_box
  loadcoords coords
##FLATBOTTOM
  if {$fb_K>0} {
    set r0 $coords($ref)
    set r1 $coords($sel)
    set dr  [vecsub $r1 $r0]
    set fx [flatbot1d [lindex $dr 0] [lindex $fb_box 0] [lindex $fb_box 1] $fb_K]
    set fy [flatbot1d [lindex $dr 1] [lindex $fb_box 2] [lindex $fb_box 3] $fb_K]
    set fz [flatbot1d [lindex $dr 2] [lindex $fb_box 4] [lindex $fb_box 5] $fb_K]
    #print "dr: $dr  fx: $fx fy: $fy fz: $fz"
    addforce $sel [list $fx $fy $fz]
  }
}
proc calcforces_endstep { } { }
''')

    def _findFiles(self, inputdir):
        # Tries to find default files if the given don't exist
        defaults = {'coordinates': ('structure.pdb',),
                    'structure': ('structure.psf', 'structure.prmtop'),
                    'parameters': ('parameters', 'structure.prmtop')}

        for field in defaults:
            userval = self.acemd.__dict__[field]
            if userval is not None and not os.path.exists(os.path.join(inputdir, userval)):
                self.acemd.__dict__[field] = None

            if self.acemd.__dict__[field] is None:
                for val in defaults[field]:
                    if os.path.exists(os.path.join(inputdir, val)):
                        self.acemd.__dict__[field] = val
                        break

            if userval is not None and self.acemd.__dict__[field] is not None and self.acemd.__dict__[field] != userval:
                logger.warning('Could not locate structure file {}. Using {} instead.'.format(
                    os.path.join(inputdir, userval), os.path.join(inputdir, self.acemd.__dict__[field])
                ))
            elif self.acemd.__dict__[field] is None:
                raise RuntimeError('Could not locate any {f:} file in {i:}. '
                                   'Please set the Equilibration.acemd.{f:} property to '
                                   'point to the {f:} file'.format(f=field, i=inputdir))

    def _amberFixes(self):
        # AMBER specific fixes
        if self.acemd.parameters.endswith('structure.prmtop'):
            self.acemd.parmfile = self.acemd.parameters
            self.acemd.parameters = None
            self.acemd.scaling14 = '0.8333333'
            self.acemd.amber = 'on'

    def write(self, inputdir, outputdir):
        """ Writes the production protocol and files into a folder.

        Parameters
        ----------
        inputdir : str
            Path to a directory containing the files produced by a equilibration process.
        outputdir : str
            Directory where to write the production setup files.
        """
        self._findFiles(inputdir)
        self._amberFixes()

        from htmd.units import convert
        numsteps = convert(self.timeunits, 'timesteps', self.runtime, timestep=self.acemd.timestep)
        self.acemd.temperature = str(self.temperature)
        self.acemd.langevintemp = str(self.temperature)
        if self.fb_k > 0: #use TCL only for flatbottom
            mol = Molecule(os.path.join(inputdir, self.acemd.coordinates))
            self.acemd.tclforces = 'on'
            tcl = list(self.acemd.TCL)
            tcl[0] = tcl[0].format(NUMSTEPS=numsteps, KCONST=self.fb_k,
                                   REFINDEX=' '.join(map(str, mol.get('index', self.fb_reference))),
                                   SELINDEX=' '.join(map(str, mol.get('index', self.fb_selection))),
                                   BOX=' '.join(map(str, self.fb_box)))
            self.acemd.TCL = tcl[0] + tcl[1]
        else:
            self.acemd.TCL = 'set numsteps {}\n'.format(numsteps)
        if self.useconstraints:
            # Turn on constraints
            self.acemd.constraints = 'on'
            self.acemd.constraintscaling = '1.0'
        else:
            if len(self.constraints) != 0:
                logger.warning('You have setup constraints to {} but constraints are turned off. '
                               'If you want to use constraints, define useconstraints=True'.format(self.constraints))
        self.acemd.setup(inputdir, outputdir, overwrite=True)

        # Adding constraints
        if self.useconstraints:
            inmol = Molecule(os.path.join(inputdir, self.acemd.coordinates))
            inmol.set('occupancy', 0)
            inmol.set('beta', 0)
            if len(self.constraints) == 0:
                raise RuntimeError('You have set the production to use constraints (useconstraints=True), but have not '
                               'defined any constraints (constraints={}).')
            else:
                for sel in self.constraints:
                    inmol.set('beta', self.constraints[sel], sel)
                outfile = os.path.join(outputdir, self.acemd.coordinates)
                inmol.write(outfile)
                self.acemd.consref = self.acemd.coordinates

if __name__ == "__main__":
    import htmd.home
    from htmd.util import tempname
    md = Production()
    md.temperature = 300
    md.fb_reference = 'protein and name CA'
    md.fb_selection = 'segname L and noh'
    md.acemd.extendedsystem = None  # use different data
    md.acemd.binindex = None  # use different data
    md.fb_box = [-20, 20, -20, 20, 43, 45]
    md.fb_k = 5
    md.write(htmd.home.home() +'/data/equilibrate', tempname())
    md.fb_k = 0
    md.write(htmd.home.home() +'/data/equilibrate', tempname())
    md.useconstraints = True
    md.constraints = {'protein and name CA': 1, 'protein and noh and not name CA': 0.1}
    md.write(htmd.home.home() +'/data/equilibrate', tempname())


# (c) 2015-2018 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
from moleculekit.molecule import Molecule
from glob import glob
import numpy as np
import logging

logger = logging.getLogger(__name__)

# Based on: http://onlinelibrary.wiley.com/doi/10.1002/(SICI)1097-0134(199601)24:1%3C92::AID-PROT7%3E3.0.CO;2-Q/epdf
# Structure, energetics, and dynamics of lipid–protein interactions: A molecular dynamics study
# of the gramicidin A channel in a DMPC bilayer


class _Lipid:
    def __init__(self, resname=None, headname=None, mol=None, xyz=None, rot=None, neighbours=None, rings=None,
                 area=None):
        self.resname = resname
        self.headname = headname
        self.mol = mol
        self.xyz = xyz
        self.rot = rot
        self.neighbours = neighbours
        self.rings = rings
        self.area = area

    def __repr__(self):
        return '<{}.{} object at {} {}>'.format(self.__class__.__module__, self.__class__.__name__, hex(id(self)),
                                                self.__str__())

    def __str__(self):
        s = ''
        if self.resname is not None:
            s += 'resname: {} '.format(self.resname)
        if self.headname is not None:
            s += 'headname: {} '.format(self.headname)
        if self.xyz is not None:
            s += 'xyz: {} '.format(self.xyz)
        if self.neighbours is not None:
            s += 'neigh: {} '.format(len(self.neighbours))
        if self.mol is not None:
            s += 'mol: {} '.format(id(self.mol))
        return s[:-1]


def listLipids():
    """ Lists all available lipids

    Examples
    --------
    >>> from htmd.membranebuilder.build_membrane import buildMembrane 
    >>> build_membrane.listLipids()
    ---- Lipids list: ...

    """

    from htmd.home import home
    import os
    from natsort import natsorted

    membranebuilderhome = os.path.join( home(shareDir=True), 'membranebuilder', 'lipids', '' )
    lipids = natsorted(glob( os.path.join(membranebuilderhome, '*', '')))
    print('---- Lipids list: ' + membranebuilderhome + ' ----')
    for l in lipids:
        print('- ', os.path.basename(os.path.abspath(l)))
    print('* Lipid DB file: ' + os.path.join(membranebuilderhome, 'lipiddb.csv'))



def _createLipids(lipidratio, area, lipiddb, files, leaflet=None):
    lipiddb = lipiddb.to_dict(orient='index')
    lipidnames = list(lipidratio.keys())
    ratiosAPL = np.array([lipidratio[lipn] * lipiddb[lipn]['APL'] for lipn in lipidnames])
    # Calculate the total areas per lipid type
    areaspl = area * (ratiosAPL / ratiosAPL.sum())
    # Calculate the counts from the total areas
    counts = np.round(areaspl / np.array([lipiddb[lipn]['APL'] for lipn in lipidnames])).astype(int)

    lipids = []
    for i in range(len(lipidnames)):
        resname = lipidnames[i]
        rings = _detectRings(Molecule(files[resname][0]))
        for k in range(counts[i]):
            if leaflet == 'upper':
                xyz = np.array([np.nan, np.nan, lipiddb[resname]['Thickness'] / 2])
            elif leaflet == 'lower':
                xyz = np.array([np.nan, np.nan, - lipiddb[resname]['Thickness'] / 2])
            lipids.append(_Lipid(resname=resname, headname=lipiddb[resname]['Head'], rings=rings, area=lipiddb[resname]['APL'], xyz=xyz))
    return lipids


def _setPositionsLJSim(width, lipids):
    from htmd.membranebuilder.ljfluid import distributeLipids

    sigmas = np.array([2 * np.sqrt(l.area / np.pi) for l in lipids])
    resnames = [l.resname for l in lipids]

    cutoff = min(np.min(width) / 2, 3 * np.max(sigmas))
    pos = distributeLipids(width + [2 * cutoff], resnames, sigmas, cutoff=cutoff)
    for i in range(len(lipids)):
        lipids[i].xyz[:2] = pos[i, :2]


def _createMembraneMolecule(lipids):
    from moleculekit.util import rotationMatrix

    allmols = []
    for i, l in enumerate(lipids):
        mol = l.mol.copy()
        headpos = mol.coords[mol.name == l.headname].flatten()[np.newaxis, :]
        mol.moveBy(-headpos)
        mol.rotateBy(rotationMatrix([0, 0, 1], np.deg2rad(l.rot)))
        mol.moveBy(l.xyz)
        mol.resid[:] = i
        allmols.append(mol)
    
    def mergeMols(mollist):  # Divide and conquer approach for merging molecules
        while len(mollist) > 1:
            mollist[0].append(mollist[1])
            mollist = [mollist[0]] + mergeMols(mollist[2:])
        if len(mollist) == 1:
            return mollist
        return []

    return mergeMols(allmols)[0]


def _detectRings(mol):
    import networkx as nx
    bonds = mol._guessBonds()

    G = nx.Graph()
    G.add_edges_from(bonds)
    cycles = np.array(nx.cycle_basis(G))
    if len(cycles) == 0:
        return None

    cyclelen = np.array([len(c) for c in cycles])
    fivesix = cycles[np.where((cyclelen == 5) | (cyclelen == 6))[0]]

    return fivesix


def wrapping_dist_python(coor1, coor2, box):
    assert (coor1.ndim == 1) or (coor2.ndim == 1)
    dist = coor1 - coor2
    dist = dist - box * np.round(dist / box)
    return np.sqrt(np.sum(dist * dist, 1))


def _findNeighbours(lipids, box):
    xypos = np.vstack([l.xyz[:2] for l in lipids])

    for i in range(len(lipids)):
        dist = wrapping_dist_python(xypos[i, :], xypos[i + 1:, :], box)
        lipids[i].neighbours = i + 1 + np.where(dist < 11)[0]


def _loadMolecules(lipids, files):
    from moleculekit.util import rotationMatrix
    # Create Molecules
    for l in lipids:
        randidx = np.random.randint(len(files[l.resname]))
        mol = Molecule(files[l.resname][randidx])
        mol.filter('not water', _logger=False)
        if l.xyz[2] < 0:
            mol.rotateBy(rotationMatrix([1, 0, 0], np.deg2rad(180)))  # Rotate the lower leaflet lipids upside down
        l.mol = mol
        l.rot = np.random.random() * 360 - 180  # Random starting rotation


def _locateLipidFiles(folder, lipidnames):
    import os
    files = {}
    for m in lipidnames:
        files[m] = glob(os.path.join(folder, m, '*.pdb'))
        if len(files[m]) == 0:
            raise RuntimeError('Could not locate pdb files for lipid "{}" in folder {}'.format(m, folder))
    return files


def buildMembrane(xysize, ratioupper, ratiolower, waterbuff=20, minimplatform='CPU', equilibrate=True, equilplatform='CUDA', outdir=None, lipidf=None):
    """ Construct a membrane containing arbitrary lipids and ratios of them.

    Parameters
    ----------
    xysize : list
        A list containing the size in x and y dimensions of the membrane in Angstroms
    ratioupper : dict
        A dict with keys the molecule names and the ratio of that molecule for the upper layer
    ratiolower : dict
        Same as ratioupper but for the lower layer
    waterbuff : float
        The z-dimension size of the water box above and below the membrane
    minimplatform : str
        The platform on which to run the minimization ('CUDA' or 'CPU')
    equilibrate : bool
        If True it equilibrates the membrane
    equilplatform : str
        The platform on which to run the equilibration ('CUDA' or 'CPU')
    outdir : str
        A folder in which to store the psf and pdb files
    lipidf : str
        The path to the folder containing the single-lipid PDB structures as well as the lipid DB file

    Returns
    -------
    mol : :class:`Molecule <moleculekit.molecule.Molecule`
        The resulting membrane including surrounding waters

    Examples
    --------
    >>> lipidratioupper = {'popc': 10, 'chl1': 1}
    >>> lipidratiolower = {'popc': 8, 'chl1': 2}
    >>> width = [50, 100]
    >>> res = buildMembrane(width, lipidratioupper, lipidratiolower)
    """
    from htmd.membranebuilder.ringpenetration import resolveRingPenetrations
    from htmd.builder.solvate import solvate
    from htmd.builder.charmm import build
    from htmd.util import tempname
    from moleculekit.molecule import Molecule
    from htmd.home import home
    import os
    import pandas as pd

    if lipidf is None:
        lipidf = os.path.join(home(shareDir=True), 'membranebuilder', 'lipids')
    lipiddb = pd.read_csv(os.path.join(lipidf, 'lipiddb.csv'), index_col='Name')

    uqlip = np.unique(list(ratioupper.keys()) + list(ratiolower.keys()))
    files = _locateLipidFiles(lipidf, uqlip)

    area = np.prod(xysize)
    lipids = _createLipids(ratioupper, area, lipiddb, files, leaflet='upper')
    lipids += _createLipids(ratiolower, area, lipiddb, files, leaflet='lower')

    _setPositionsLJSim(xysize, [l for l in lipids if l.xyz[2] > 0])
    _setPositionsLJSim(xysize, [l for l in lipids if l.xyz[2] < 0])

    _findNeighbours(lipids, xysize)

    _loadMolecules(lipids, files)

    # from globalminimization import minimize
    # newpos, newrot = minimize(lipids, xysize + [100], stepxy=0.5, steprot=50, contactthresh=1)
    # for i in range(len(lipids)):
    #     lipids[i].xyz[:2] = newpos[i]
    #     lipids[i].rot = newrot[i]

    resolveRingPenetrations(lipids, xysize)
    memb = _createMembraneMolecule(lipids)

    minc = memb.get('coords', 'name P').min(axis=0) - 5
    maxc = memb.get('coords', 'name P').max(axis=0) + 5

    mm = [[0, 0, maxc[2] - 2], [xysize[0], xysize[1], maxc[2] + waterbuff]]
    smemb = solvate(memb, minmax=mm)
    mm = [[0, 0, minc[2] - waterbuff], [xysize[0], xysize[1], minc[2] + 2]]
    smemb = solvate(smemb, minmax=mm)

    smemb.moveBy([0, 0, -smemb.coords[:, 2, 0].min()])

    if outdir is None:
        outdir = tempname()
        logger.info('Outdir {}'.format(outdir))
    res = build(smemb, ionize=False, stream=['str/lipid/toppar_all36_lipid_cholesterol_model_1.str'], outdir=outdir)

    if equilibrate:
        from htmd.membranebuilder.simulate_openmm import equilibrateSystem
        from shutil import copy, move
        outpdb = tempname(suffix='.pdb')
        charmmf = os.path.join(home(shareDir=True), 'membranebuilder', 'charmm-toppar')
        equilibrateSystem(os.path.join(outdir, 'structure.pdb'), os.path.join(outdir, 'structure.psf'), outpdb,
                          charmmfolder=charmmf, equilplatform=equilplatform, minimplatform=minimplatform)
        res = Molecule(outpdb)
        res.center()
        move(os.path.join(outdir, 'structure.pdb'), os.path.join(outdir, 'starting_structure.pdb'))
        copy(outpdb, os.path.join(outdir, 'structure.pdb'))

    return res


def _findLeastAreaLipid(folder):
    """
    Use this to select the single least stretched conformation from a CHARMM-GUI lipid library
    """
    from glob import glob
    from scipy.spatial.distance import cdist
    ff = glob(folder + '/*/*.crd')
    maxdist = []
    for f in ff:
        m = Molecule(f)
        center = m.coords.mean(axis=0)
        dists = cdist(m.coords[:, :2, 0], center[:2].T)
        maxdist.append(dists.max())
    return ff[np.argmin(maxdist)], np.min(maxdist)


if __name__ == '__main__':
    pass






# (c) 2015-2018 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
import numpy as np
import scipy.spatial.distance as distance
from moleculekit.molecule import Molecule
from moleculekit.util import sequenceID
import logging
logger = logging.getLogger(__name__)

# Info about all ions, with charge, name, amber and charmm names (may
# need residues too). Only the first column is currently used.
_ions = {
    'NA': (1, 'sodium', 'Na+', 'SOD'),
    'MG': (2, 'magnesium', 'Mg2+', 'MG'),
    'ZN': (2, 'zinc', 'Zn2+', 'ZN'),
    'K': (1, 'potassium', 'K+', 'POT'),
    'CS': (1, 'cesium', 'Cs+', 'CES'),
    'CA': (2, 'calcium', 'Ca2+', 'CAL'),
    'CL': (-1, 'chloride', 'Cl-', 'CLA')
}


def ionize(netcharge, nwater, neutralize=True, saltconc=None, cation=None, anion=None, ff=None):
    if ff:
        logger.warning('The ff option is no longer needed and is deprecated. '
                       'Please use standard names (e.g. NA, CL, K) for `cation` and `anion` arguments.')
    
    if cation is None:
            cation = 'NA'
    if anion is None:
            anion = 'CL'

    if saltconc is not None:
        neutralize = False

    cationcharge = _ionGetCharge(cation)
    anioncharge = _ionGetCharge(anion)

    roundnetcharge = int(np.round(netcharge))
    done = False
    ncation = 0
    nanion = 0

    # First we neutralize the system
    if netcharge > 0:
        if anioncharge == -1:
            nanion = roundnetcharge
            ncation = 0
        elif anioncharge == -2:
            if np.mod(roundnetcharge, 2) == 0:  # Total charge is even
                nanion = int(roundnetcharge / 2)
                ncation = 0
            else:  # The charge is odd
                if cationcharge == 1:
                    nanion = 1 + int(np.floor(roundnetcharge / 2))
                    ncation = 1
                else:
                    raise NameError('Failed to neutralize because the cation and anion charges cannot reach 0')
    elif netcharge < 0:
        roundnetchargeabs = abs(roundnetcharge)
        if cationcharge == 1:
            nanion = 0
            ncation = roundnetchargeabs
        elif cationcharge == 2:
            if np.mod(roundnetchargeabs, 2) == 0:
                nanion = 0
                ncation = int(roundnetchargeabs / 2)
            else:
                if anioncharge == -1:
                    nanion = 1
                    ncation = 1 + int(np.floor(roundnetchargeabs / 2))
                else:
                    raise NameError('Failed to neutralize because the cation and anion charges cannot reach 0')
    else:
        logger.info('The system is already neutral.')
        done = True

    if not done:
        newtotalcharge = nanion * anioncharge + ncation * cationcharge + roundnetcharge
        if newtotalcharge != 0:
            raise NameError('Could not neutralize the system')
    if neutralize:
        return anion, cation, nanion, ncation

    # If we want a specific salt concentration, guess chemical formula
    if cationcharge == 1 and anioncharge == -1:  # e.g., NaCl, Kcl, ...
        cationstoich = 1
        anionstoich = 1
    elif cationcharge == 2 and anioncharge == -1:  # e.g., MgCl2
        cationstoich = 1
        anionstoich = 2
    elif cationcharge == 1 and anioncharge == -2:
        cationstoich = 2
        anionstoich = 1
    elif cationcharge == 2 and anioncharge == -2:
        cationstoich = 1
        anionstoich = 1
    else:
        raise NameError('Unsupported ion charge; cannot guess chamical formula.')

    num = int(np.floor(0.5 + 0.0187 * saltconc * nwater))
    logger.info('Adding {} anions + {} cations for neutralizing and {} ions for the given salt concentration.'.format(nanion, ncation, (cationstoich + anionstoich) * num))
    ncation += cationstoich * num
    nanion += anionstoich * num
    return anion, cation, anion, cation, nanion, ncation


def _ionGetCharge(ion):
    if ion not in _ions:
        raise NameError('Ion {:s} not in the database'.format(ion))
    return _ions[ion][0]


def ionizePlace(mol, anion_resname, cation_resname, anion_name, cation_name, nanion, ncation, dfrom=5, dbetween=5, segname=None):
    """Place a given number of negative and positive ions in the solvent.

    Replaces water molecules al long as they respect the given distance criteria.

    Parameters
    ----------
    mol : :class:`Molecule <moleculekit.molecule.Molecule>` object
        The Molecule object
    anion_resname : str
        Resname of the added anions
    cation_resname : str
        Resname of the added cations
    anion_name : str
        Name of the added anions
    cation_name : str
        Name of the added cations
    nanion : int
        Number of anions to add
    ncation : int
        Number of cations to add
    dfrom : float
        Min distance of ions from molecule
    dbetween : float
        Min distance between ions
    segname : str
        Segment name to add
        
    Returns
    -------
    mol : :class:`Molecule <moleculekit.molecule.Molecule>` object
        The molecule with the ions added
    """

    newmol = mol.copy()

    logger.debug('Min distance of ions from molecule: ' + str(dfrom) + 'A')
    logger.debug('Min distance between ions: ' + str(dbetween) + 'A')
    logger.debug('Placing {:d} anions and {:d} cations.'.format(nanion,ncation))

    if (nanion + ncation) == 0:
        return newmol

    nions = nanion + ncation

    betabackup = newmol.beta.copy()
    newmol.set('beta', sequenceID((newmol.resid, newmol.insertion, newmol.segid)))

    # Find water oxygens to replace with ions
    ntries = 0
    maxtries = 10
    while True:
        ionlist = []
        watindex = newmol.atomselect('noh and water and not (within ' + str(dfrom) + ' of not water)', indexes=True)
        watsize = len(watindex)

        if watsize == 0:
            raise NameError('No waters could be found further than ' + str(dfrom) + ' from other molecules to be replaced by ions. You might need to solvate with a bigger box or disable the ionize property when building.')

        while len(ionlist) < nions:
            if len(watindex) == 0:
                break
            randwat = np.random.randint(len(watindex))
            thision = watindex[randwat]
            addit = True
            if len(ionlist) != 0:  # Check for distance from precious ions
                ionspos = newmol.get('coords', sel=ionlist)
                thispos = newmol.get('coords', sel=thision)
                dists = distance.cdist(np.atleast_2d(ionspos), np.atleast_2d(thispos), metric='euclidean')

                if np.any(dists < dbetween):
                    addit = False
            if addit:
                ionlist.append(thision)
                watindex = np.delete(watindex, randwat)
        if len(ionlist) == nions:
            break

        ntries += 1
        if ntries == maxtries:
            raise NameError('Failed to add ions after ' + str(maxtries) + ' attempts. Try decreasing the ''from'' and ''between'' parameters, decreasing ion concentration or making a larger water box.')

    # Delete waters but keep their coordinates
    waterpos = np.atleast_2d(newmol.get('coords', ionlist))
    betasel = np.zeros(newmol.numAtoms, dtype=bool)
    for b in newmol.beta[ionlist]:
        betasel |= newmol.beta == b
    atmrem = np.sum(betasel)
    atmput = 3 * len(ionlist)
    # assert atmrem == atmput, 'Removing {} atoms instead of {}. Report this bug.'.format(atmrem, atmput)
    sel = np.where(betasel)[0]
    newmol.remove(sel, _logger=False)
    # assert np.size(sel) == atmput, 'Removed {} atoms instead of {}. Report this bug.'.format(np.size(sel), atmput)
    betabackup = np.delete(betabackup, sel)

    # Add the ions
    randidx = np.random.permutation(np.size(waterpos, 0))
    atom = Molecule()
    atom.empty(1)
    atom.set('chain', 'I')
    atom.set('segid', 'I')

    for i in range(nanion):
        atom.set('name', anion_name)
        atom.set('resname', anion_resname)
        atom.set('resid', newmol.resid[-1] + 1)
        atom.coords = waterpos[randidx[i], :]
        newmol.insert(atom, len(newmol.name))
    for i in range(ncation):
        atom.set('name', cation_name)
        atom.set('resname', cation_resname)
        atom.set('resid', newmol.resid[-1] + 1)
        atom.coords = waterpos[randidx[i+nanion], :]
        newmol.insert(atom, len(newmol.name))

    # Restoring the original betas
    newmol.beta[:len(betabackup)] = betabackup
    return newmol


def _getSegname(mol, segname):
    defsegnames = ['ION','ION1','ION3','ION4','ION5','ION6','ION7','ION8','ION9','IN1','IN2','IN3','IN4','IN5','IN6','IN7','IN8','IN9']

    # Make sure segname is not taken
    if segname is None:
        for i in range(len(defsegnames)):
            sel = mol.segid == defsegnames[i]
            if not np.any(sel):
                segname = defsegnames[i]
                break
        if segname is None:
            raise NameError('All default segnames taken! Provide your own.')
    else:
        sel = mol.segid == segname
        if np.any(sel):
            raise NameError('Segname already taken. Provide a different one.')



# (c) 2015-2018 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#

# Rapidly exploring random tree to find reaction coordinate for umbrella sampling!
# Use VMD draw line to visualize! (dashed for failed, solid for connected)
import numpy as np
from moleculekit.util import maxDistance, uniformRandomRotation
from moleculekit.molecule import Molecule
from moleculekit.vmdviewer import getCurrentViewer
from scipy.spatial.distance import cdist


class Tree:
    def __init__(self, initialpoints):
        if isinstance(initialpoints, np.ndarray):
            initialpoints = initialpoints.tolist()
        self.points = initialpoints
        self.parent = [None] * len(self.points)
        self.costs = [0] * len(self.points)

    def addPoint(self, point, parent, dist):
        if isinstance(point, np.ndarray):
            point = point.tolist()
        if isinstance(dist, np.ndarray):
            dist = dist[0]
        self.points.append(point)
        self.parent.append(parent)
        self.costs.append(self.costs[parent] + dist)


def _getCoordinates(mol, ligandsel, ligcom):
    mol.center(sel='protein')
    protcoor = mol.get('coords', sel='protein and not {}'.format(ligandsel))
    if ligcom:
        ligcoor = np.mean(mol.get('coords', sel=ligandsel), axis=0)[np.newaxis, :]
    else:
        ligcoor = mol.get('coords', sel=ligandsel)
    return protcoor, ligcoor


def _newPoint(p_target, p_begin, step):
    v = p_target - p_begin
    norm_v = v / np.linalg.norm(v)
    if isinstance(step, np.ndarray):
        return p_begin + (step[:, np.newaxis] * norm_v)
    else:
        return p_begin + (step * norm_v)


def _dist(a, b):
    if isinstance(a, list):
        a = np.array(a)
    if isinstance(b, list):
        b = np.array(b)
    return cdist(np.atleast_2d(a), np.atleast_2d(b))


def _getNearest(coords, point):
    dist = _dist(coords, point)
    idx = np.argmin(dist)
    return idx, coords[idx]


def _collision(coords, point, buffer=2):
    dist = _dist(coords, point)
    if np.any(dist < buffer):
        return True
    else:
        return False


def _pointsOnSphere(radius, numsamples=1000):
    pointcoords = np.zeros((numsamples, 3))
    for i in range(numsamples):
        point = Molecule()
        point.empty(1)
        point.moveBy([0, 0, radius])
        point.rotateBy(uniformRandomRotation())
        pointcoords[i, :] = np.squeeze(point.coords)

    return pointcoords


def _collisionFreePath(obstacles, start, stop, colldist, step):
    linepoints = _newPoint(stop, start, np.arange(step, _dist(start, stop), step=step))
    distances = _dist(linepoints, obstacles)
    if np.any(distances <= colldist):
        return False
    else:
        return True


def _collisionFreeNeighbours(tree, point, radius, obstacles, colldist, step):
    dists = _dist(tree.points, point)
    near = np.where(dists < radius)[0]
    nn = []
    for n in near:
        if _collisionFreePath(obstacles, point, tree.points[n], colldist, step):
            nn.append(n)
    near = nn
    return near, dists[near]


def _rewire(tree, near, neardist, newidx):
    # Remove parent from near neighbours
    parentidx = np.where(near == tree.parent[newidx])[0]
    near = np.delete(near, parentidx)
    neardist = np.delete(neardist, parentidx)
    if len(near) == 0:
        return

    costs = np.array(tree.costs)
    newcosts = np.squeeze(neardist) + costs[newidx]
    shorter = np.where(newcosts < costs[near])[0]
    for s in shorter:
        #from IPython.core.debugger import Tracer
        #Tracer()()
        tree.costs[near[s]] = newcosts[s]
        tree.parent[near[s]] = newidx


def _chooseParent(tree, near, neardist):
    if len(near) == 0:
        raise AssertionError('Should include at least p_near.')

    costs = np.array(tree.costs)
    closecosts = costs[near] + np.squeeze(neardist)
    shortest = np.argmin(closecosts)
    return near[shortest], neardist[shortest]


def _endCondition(protcoor, target, curr, outdist, method='target'):
    if method == 'target':
        return _dist(target, curr) < 2
    if method == 'exited':
        return np.all(_dist(protcoor, curr) >= outdist)


def _pathOptimize(tree, currnode, obstacles, colldist, step):
    start = np.array(tree.points[currnode])
    endnode = tree.parent[currnode]
    prevend = endnode

    while True:
        if endnode is None:
            tree.parent[currnode] = prevend
            tree.costs[currnode] = tree.costs[prevend] + _dist(tree.points[currnode], tree.points[prevend])
            return

        stop = np.array(tree.points[endnode])
        if _collisionFreePath(obstacles, start, stop, colldist, step):
            prevend = endnode
            endnode = tree.parent[endnode]
        else:
            tree.parent[currnode] = prevend
            tree.costs[currnode] = tree.costs[prevend] + _dist(tree.points[currnode], tree.points[prevend])
            _pathOptimize(tree, prevend, obstacles, colldist, step)
            return


def _getBeacons(tree, endnode):
    beacons = []
    currnode = tree.parent[endnode]
    while tree.parent[currnode] is not None:
        beacons.append(currnode)
        currnode = tree.parent[currnode]
    return beacons


def _randomPoint(min, max):
    lens = max - min
    return np.random.rand(3) * lens + min


def rrtstarsmart(mol, ligandsel, step=1, maxiter=int(1E6), ligcom=False, colldist=2, outdist=8, radius=2.5):
    protcoor, ligcoor = _getCoordinates(mol, ligandsel, ligcom)
    viewer = _prepareViewer(mol, ligandsel)

    mincoor = np.squeeze(np.min(mol.coords, axis=0) - 10)
    maxcoor = np.squeeze(np.max(mol.coords, axis=0) + 10)
    print(mincoor, maxcoor)

    spherecoor = _pointsOnSphere(maxDistance(mol) + 5)

    tree = Tree(ligcoor)
    nocol = None

    for i in range(maxiter):
        print(i)
        #p_rand = spherecoor[np.random.randint(spherecoor.shape[0]), :]
        p_rand = _randomPoint(mincoor, maxcoor)
        parent_idx, p_near = _getNearest(tree.points, p_rand)
        p_new = _newPoint(p_rand, p_near, step)

        if _collision(protcoor, p_new, buffer=colldist):
            continue

        print('no collision')
        near, neardist = _collisionFreeNeighbours(tree, p_new, radius, protcoor, colldist, step)
        parent_idx, dist = _chooseParent(tree, near, neardist)
        tree.addPoint(p_new, parent_idx, dist)
        _rewire(tree, near, neardist, len(tree.points)-1)

        if _endCondition(protcoor, p_rand, p_new, outdist, method='exited'):
            break

    for i in range(len(tree.points)):
        if tree.parent[i] is not None:
            _drawline(viewer, tree.points[i], tree.points[tree.parent[i]])

    print('optimizing')
    _pathOptimize(tree, len(tree.points)-1, protcoor, colldist, step)
    print('finished optimization')
    # TODO: The algorithm is not finished. Normally you would get beacons and continue sampling around them to optimize
    # TODO: the path even more. However, we don't really need this in our case, unless I decide to make a real full impl
    # TODO: http://ieeexplore.ieee.org/xpls/icp.jsp?arnumber=6284384

    currnode = len(tree.points)-1
    viewer.send('draw color green')
    while tree.parent[currnode] is not None:
        _drawline(viewer, tree.points[currnode], tree.points[tree.parent[currnode]])
        currnode = tree.parent[currnode]

    return


def rrt(mol, ligandsel, step=1, maxiter=int(1E6), ligcom=False, colldist=2, outdist=8):
    protcoor, ligcoor = _getCoordinates(mol, ligandsel, ligcom)
    viewer = _prepareViewer(mol, ligandsel)

    spherecoor = _pointsOnSphere(maxDistance(mol) + 5)

    tree = Tree(ligcoor)
    nocol = None

    for i in range(maxiter):
        print(i)
        if nocol is None:
            p_rand = spherecoor[np.random.randint(spherecoor.shape[0]), :]  # TODO: Make it clever. Discard failed points
        else:  # Reusing the same point that caused no collision
            p_rand = nocol
        parent_idx, p_near = _getNearest(tree.points, p_rand)
        p_new = _newPoint(p_rand, p_near, step)

        if _collision(protcoor, p_new, buffer=colldist):
            nocol = None
            continue
        nocol = p_rand

        print('no collision')
        tree.addPoint(p_new, parent_idx, step)

        _drawline(viewer, p_near, p_new)
        if _dist(p_rand, p_new) < 2:
            break


def raytracing(mol, ligandsel, step=1, colldist=2, outdist=8, ligcom=False, numsamples=2000, ratioexposed=0, vmd=True):
    protcoor, ligcoor = _getCoordinates(mol, ligandsel, ligcom)

    spherecoor = _pointsOnSphere(maxDistance(mol) + 5, numsamples=numsamples)
    if vmd:
        _viewSphere(spherecoor)
        viewer = _prepareViewer(mol, ligandsel)

    distances = _dist(ligcoor, spherecoor)

    from joblib import Parallel, delayed
    results = Parallel(n_jobs=-2, verbose=11)(
        delayed(parallelfunc)(j, spherecoor[j, :], ligcoor, protcoor, step, colldist, outdist, distances[:, j])
        for j in range(spherecoor.shape[0]))

    points = []
    pointdist = []
    shortpoints = []
    for r in results:
        points += r[0]
        pointdist += r[1]
        shortpoints += r[2]

    if len(points) == 0:
        raise RuntimeError('No ligand atoms can exit the pocket without {}A clashes.'.format(colldist))

    points = np.array(points)
    numexposed = len(np.unique(points[:, 0]))
    percentexposed = (numexposed / ligcoor.shape[0]) * 100
    if numexposed < (ratioexposed * ligcoor.shape[0]):
        raise RuntimeError('Only {:.1f}% ligand atoms can exit the pocket without {}A clashes. '
                           'This collides with the user-defined required {:.1f}% exposed ligand atoms.'.format(
                            percentexposed, colldist, ratioexposed*100))
    print('{:.1f}% ligand atoms can exit the pocket without {}A clashes.'.format(percentexposed, colldist))


    from scipy.stats import mode
    modesphere = mode(points[:, 1]).mode[0]
    idx = points[:, 1] == modesphere
    meanlig = np.mean(ligcoor[points[idx, 0], :], axis=0)

    if vmd:
        viewer.send('draw color green')
        _drawline(viewer, meanlig, spherecoor[modesphere, :])

        for idx, p in enumerate(points.tolist()):
            if p[1] == modesphere:
                viewer.send('draw color red')
                _drawline(viewer, shortpoints[idx], spherecoor[p[1], :])
                viewer.send('draw color green')
                _drawline(viewer, ligcoor[p[0], :], shortpoints[idx])

    return spherecoor[modesphere, :] - meanlig


def parallelfunc(j, spherecoor, ligcoor, protcoor, step, colldist, outdist, distances):
    points = []
    pointdist = []
    shortpoints = []
    for i in range(ligcoor.shape[0]):
        linepoints = _newPoint(spherecoor, ligcoor[i, :], np.arange(step, distances[i], step=step))
        collisions = _dist(linepoints, protcoor)
        if not np.any(collisions <= colldist):
            linep_mincoll = np.min(collisions, axis=1)
            idx = np.min(np.where(linep_mincoll >= outdist)[0])
            points.append([i, j])
            pointdist.append(step * (idx+1))
            shortpoints.append(linepoints[idx, :])
    return points, pointdist, shortpoints


def _drawline(viewer, start, end, dashed=True):
    if dashed:
        append = 'style dashed'
    else:
        append = ''
    viewer.send('draw line {{ {} }} {{ {} }} {}'.format(' '.join(map(str, start)), ' '.join(map(str, end)), append))


def _prepareViewer(mol, ligandsel):
    mol.view(sel='protein', style='lines', hold=True)
    mol.view(sel=ligandsel, style='licorice', hold=True)
    mol.view()

    viewer = getCurrentViewer()
    viewer.send('draw color red')
    viewer.send('color Display Background black')
    viewer.send('draw materials off')
    return viewer


def _viewSphere(spherecoor):
    spheremol = Molecule()
    spheremol.empty(spherecoor.shape[0])
    spheremol.coords = np.atleast_3d(spherecoor)
    spheremol.view(guessBonds=False)



# (c) 2015-2018 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
from htmd.apps.acemd import Acemd as Acemd2
from htmd.mdengine.acemd.acemd import Acemd, _Restraint, GroupRestraint, AtomRestraint
from htmd.config import _config
from protocolinterface import ProtocolInterface, val
import os
import numpy as np
import logging

logger = logging.getLogger(__name__)


class Equilibration(ProtocolInterface):
    """ Equilibration protocol v3

        Equilibration protocol for globular and membrane proteins
        Supporst extra restraints like a flatbottom potential box to retain a ligand
        for example within this box.

        Parameters
        ----------
        runtime : float, default=0
            Running time of the simulation.
        timeunits : str, default='steps'
            Units for time arguments. Can be 'steps', 'ns' etc.
        temperature : float, default=300
            Temperature of the thermostat in Kelvin
        restraints : list, default=None
            A list of restraint objects. See :class:`AtomRestraint <htmd.mdengine.acemd.acemd.AtomRestraint>` and:class:`GroupRestraint<htmd.mdengine.acemd.acemd.GroupRestraint>`)
        useconstantratio : bool, default=False
            For membrane protein simulations set it to true so that the barostat does not modify the xy aspect ratio.
        restraintsteps : int, default=None
            Number of initial steps to apply restraints in units of 4fs. Defaults to half the simulation time.

        Example
        -------
        >>> from htmd.protocols.equilibration_v3 import Equilibration
        >>> from htmd.mdengine.acemd.acemd import GroupRestraint
        >>> md = Equilibration()
        >>> md.runtime = 4
        >>> md.timeunits = 'ns'
        >>> md.temperature = 300
        >>> md.useconstantratio = True  # only for membrane sims
        Use a 10A flat bottom potential to prevent the ligand from diffusing from original position during equilibration
        >>> width = np.array([10, 10, 10])
        >>> flatbot = GroupRestraint('segname L and noh', width, [(5, '0ns')])
        Add also the default restraints for protein
        >>> md.restraints = [flatbot,] + md.defaultEquilRestraints('2ns')
        >>> md.write('./build','./equil')
    """

    def __init__(self, _version=_config["acemdversion"]):
        if _version == 2:
            raise RuntimeError("Equilibration v3 only supports _version=3")

        super().__init__()
        self._arg(
            "acemd",
            ":class:`Acemd <htmd.mdengine.acemd.acemd.Acemd>` object",
            "Acemd class object",
            None,
            val.Object([Acemd2, Acemd]),
        )
        self._arg(
            "runtime",
            "float",
            "Running time of the simulation.",
            25000,
            val.Number(float, "0POS"),
        )
        self._arg(
            "timeunits",
            "str",
            "Units for time arguments. Can be 'steps', 'ns' etc.",
            "steps",
            val.String(),
        )
        self._arg(
            "temperature",
            "float",
            "Temperature of the thermostat in Kelvin",
            300,
            val.Number(float, "ANY"),
        )
        self._arg(
            "useconstantratio",
            "bool",
            "For membrane protein simulations set it to true so that the barostat "
            "does not modify the xy aspect ratio.",
            False,
            val.Boolean(),
        )
        self._arg(
            "restraintsteps",
            "int",
            "Number of initial steps to apply restraints in units of 4fs. Defaults "
            "to half the simulation time.",
            None,
            val.Number(int, "ANY"),
        )
        self._arg(
            "restraints",
            "list",
            "A list of restraint objects. If None will apply defaultEquilRestraints decaying over half the runtime. If no restraints are required set to empty list []."
            "See :class:`AtomRestraint <htmd.mdengine.acemd.acemd.AtomRestraint>` and"
            ":class:`GroupRestraint <htmd.mdengine.acemd.acemd.GroupRestraint>`)",
            None,
            val.Object(_Restraint),
            nargs="*",
        )

        self.acemd = Acemd()
        self.acemd.coordinates = None
        self.acemd.structure = None
        self.acemd.parameters = None
        self.acemd.restart = "on"
        self.acemd.trajectoryfile = "output.xtc"
        self.acemd.trajectoryperiod = 25000
        self.acemd.timestep = 4
        self.acemd.switching = "on"
        self.acemd.switchdistance = 7.5
        self.acemd.cutoff = 9
        self.acemd.thermostat = "on"
        self.acemd.thermostatdamping = 1
        self.acemd.pme = "on"
        self.acemd.barostat = "on"
        self.acemd.barostatpressure = 1.01325
        self.acemd.minimize = 500

    def _findFiles(self, inputdir):
        # Tries to find default files if the given don't exist
        defaults = {
            "coordinates": ("structure.pdb",),
            "structure": ("structure.psf", "structure.prmtop"),
            "parameters": ("parameters", "structure.prmtop"),
        }

        for field in defaults:
            userval = self.acemd.__dict__[field]
            if userval is not None and not os.path.exists(
                os.path.join(inputdir, userval)
            ):
                self.acemd.__dict__[field] = None

            if self.acemd.__dict__[field] is None:
                for val in defaults[field]:
                    if os.path.exists(os.path.join(inputdir, val)):
                        self.acemd.__dict__[field] = val
                        break

            if (
                userval is not None
                and self.acemd.__dict__[field] is not None
                and self.acemd.__dict__[field] != userval
            ):
                logger.warning(
                    "Could not locate structure file {}. Using {} instead.".format(
                        os.path.join(inputdir, userval),
                        os.path.join(inputdir, self.acemd.__dict__[field]),
                    )
                )
            elif self.acemd.__dict__[field] is None:
                raise RuntimeError(
                    "Could not locate any {f:} file in {i:}. "
                    "Please set the {name:}.acemd.{f:} property to "
                    "point to the {f:} file".format(
                        f=field, i=inputdir, name=self.__class__.__name__
                    )
                )

    def _amberFixes(self):
        # AMBER specific fixes
        if self.acemd.parameters.endswith("structure.prmtop"):
            self.acemd.parmfile = self.acemd.parameters
            self.acemd.parameters = None

    def defaultEquilRestraints(self, decay):
        """ Get the default protein restraints

        Parameters
        ----------
        decay : str
            The restrains will get scaled to 0 over this much time.

        Returns
        -------
        restraints : list
            A list of default protein restraints

        Examples
        --------
        >>> md = Equilibration()
        >>> res = md.defaultEquilRestraints('20ns')
        """
        caatoms = AtomRestraint("protein and name CA", 0, [(1, 0), (0, decay)])
        notcaatoms = AtomRestraint(
            "protein and noh and not name CA", 0, [(0.1, 0), (0, decay)]
        )
        return [caatoms, notcaatoms]

    def write(self, inputdir, outputdir):
        """ Write the equilibration protocol

        Writes the equilibration protocol and files into a folder for execution
        using files inside the inputdir directory

        Parameters
        ----------
        inputdir : str
            Path to a directory containing the files produced by a build process.
        outputdir : str
            Directory where to write the equilibration setup files.

        Examples
        --------
        >>> md = Equilibration()
        >>> md.write('./build','./equil')
        """

        from moleculekit.molecule import Molecule

        self._findFiles(inputdir)
        self._amberFixes()

        from htmd.units import convert

        numsteps = convert(
            self.timeunits, "timesteps", self.runtime, timestep=self.acemd.timestep
        )

        self.acemd.temperature = self.temperature
        self.acemd.thermostattemperature = self.temperature
        self.acemd.run = str(numsteps)

        pdbfile = os.path.join(inputdir, self.acemd.coordinates)
        inmol = Molecule(pdbfile)

        from htmd.builder.builder import detectCisPeptideBonds

        detectCisPeptideBonds(inmol)

        if np.any(inmol.atomselect("lipids")) and not self.useconstantratio:
            logger.warning(
                "Lipids detected in input structure. We highly recommend setting useconstantratio=True "
                "for membrane simulations."
            )

        if self.restraintsteps is None:
            constrsteps = int(numsteps / 2)
        else:
            constrsteps = int(self.restraintsteps)

        if self.restraints is not None:
            self.acemd.restraints = self.restraints
        else:
            self.acemd.restraints = self.defaultEquilRestraints(constrsteps)

        if self.acemd.boxsize is None and self.acemd.extendedsystem is None:
            coords = inmol.get("coords", sel="water")
            if coords.size == 0:  # It's a vacuum simulation
                coords = inmol.get("coords", sel="all")
                dim = np.max(coords, axis=0) - np.min(coords, axis=0)
                dim += 12.0
            else:
                dim = np.max(coords, axis=0) - np.min(coords, axis=0)
            self.acemd.boxsize = "{} {} {}".format(dim[0], dim[1], dim[2])

        if self.useconstantratio:
            self.acemd.useconstantratio = "on"

        self.acemd.setup(inputdir, outputdir, overwrite=True)


import unittest


class _TestEquilibration(unittest.TestCase):
    def _cutfirstline(self, infile, outfile):
        # Cut out the first line of prmtop which has a build date in it
        with open(infile, "r") as fin:
            data = fin.read().splitlines(True)
        with open(outfile, "w") as fout:
            fout.writelines(data[1:])

    def _compareResultFolders(self, compare, tmpdir, pid):
        from glob import glob
        import os
        import filecmp

        ignore_ftypes = (".log", ".txt")
        files = []
        deletefiles = []
        for f in glob(os.path.join(compare, "*")):
            fname = os.path.basename(f)
            if os.path.splitext(f)[1] in ignore_ftypes:
                continue
            if f.endswith("prmtop"):
                self._cutfirstline(f, os.path.join(compare, fname + ".mod"))
                self._cutfirstline(
                    os.path.join(tmpdir, fname), os.path.join(tmpdir, fname + ".mod")
                )
                files.append(os.path.basename(f) + ".mod")
                deletefiles.append(os.path.join(compare, fname + ".mod"))
            else:
                files.append(os.path.basename(f))

        match, mismatch, error = filecmp.cmpfiles(tmpdir, compare, files, shallow=False)
        if len(mismatch) != 0 or len(error) != 0 or len(match) != len(files):
            raise RuntimeError(
                "Different results produced by amber.build for test {} between {} and {} in files {}.".format(
                    pid, compare, tmpdir, mismatch
                )
            )

        for f in deletefiles:
            os.remove(f)

    def test_acemd3(self):
        from htmd.util import tempname
        from htmd.home import home
        from htmd.units import convert
        from glob import glob
        import os
        from htmd.mdengine.acemd.acemd import GroupRestraint

        eq = Equilibration()
        eq.runtime = 4
        eq.timeunits = "ns"
        eq.temperature = 300
        eq.restraintsteps = convert(
            eq.timeunits, "timesteps", eq.runtime, timestep=eq.acemd.timestep
        )
        ligres = GroupRestraint(
            "resname MOL and noh",
            [42, 38, 1],
            [(5, 0)],
            fbcentre=[-0.178, -0.178, 29.195],
        )
        eq.restraints = eq.defaultEquilRestraints(1000000) + [ligres]
        tmpdir = tempname()
        eq.write(
            home(dataDir=os.path.join("test-protocols", "build", "protLig")), tmpdir
        )

        # Compare with reference
        refdir = home(
            dataDir=os.path.join(
                "test-protocols", "equilibration", "acemd3", "protLig", "prerun"
            )
        )
        files = [os.path.basename(f) for f in glob(os.path.join(refdir, "*"))]
        self._compareResultFolders(refdir, tmpdir, "protLig")

    @unittest.skipUnless("ACE3ARG" in os.environ, "Untrusted PR")
    def test_run_water(self):
        from htmd.util import tempname
        from htmd.home import home
        from glob import glob
        import subprocess
        from subprocess import check_output
        import shutil
        import os

        acemd3exe = shutil.which("acemd3", mode=os.X_OK)
        if not acemd3exe:
            raise NameError(
                "Could not find acemd3, or no execute permissions are given"
            )

        for system in ["amber-build", "charmm-build"]:
            eq = Equilibration()
            eq.runtime = 5
            eq.timeunits = "steps"
            eq.temperature = 300
            eq.restraints = []
            # Set these down for tiny box size of water
            eq.acemd.cutoff = 3
            eq.acemd.switchdistance = 2
            eq.acemd.minimize = 50  # Do few steps to finish fast
            ######
            tmpdir = tempname()
            eq.write(
                home(dataDir=os.path.join("test-acemd", "tiny-water", system)), tmpdir
            )
            print(tmpdir)
            try:
                res = check_output(
                    ["acemd3", "--platform", "CPU", os.getenv("ACE3ARG")], cwd=tmpdir
                )
            except subprocess.CalledProcessError as exc:
                assert (
                    False
                ), f'Failed to run due to error: {exc}\n\n ---> Error log:\n\n{exc.output.decode("ascii")}'
            res = res.decode("utf-8").strip()
            assert "Completed minimization" in res, "Failed at system " + system
            assert res.endswith("Completed simulation!"), "Failed at system " + system


if __name__ == "__main__":
    unittest.main(verbosity=2)



from jobqueues.lsfqueue import LsfQueue
import logging


logger = logging.getLogger(__name__)

logger.warning('Please do not import from htmd.queues.lsfqueue. This will be deprecated. '
                'Use jobqueues.lsfqueue instead.')


from jobqueues.localqueue import LocalCPUQueue, LocalGPUQueue
import logging


logger = logging.getLogger(__name__)

logger.warning('Please do not import from htmd.queues.localqueue. This will be deprecated. '
                'Use jobqueues.localqueue instead.')


# (c) 2015-2018 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
from __future__ import print_function


from htmd.home import home
import numpy as np
import os
from os.path import join
from glob import glob
from moleculekit.util import _missingSegID, sequenceID
import shutil
from htmd.builder.builder import (
    detectDisulfideBonds,
    detectCisPeptideBonds,
    convertDisulfide,
    _checkMixedSegment,
    BuildError,
    UnknownResidueError,
    MissingAngleError,
    MissingBondError,
    MissingParameterError,
    MissingTorsionError,
    MissingAtomTypeError,
)
from subprocess import call, check_output, DEVNULL
from moleculekit.molecule import Molecule
from htmd.builder.ionize import ionize as ionizef, ionizePlace
from htmd.util import ensurelist
import unittest

import logging

logger = logging.getLogger(__name__)


def _findTeLeap():
    teleap = shutil.which("teLeap", mode=os.X_OK)
    if not teleap:
        raise FileNotFoundError(
            "teLeap not found. You should either have AmberTools or ambermini installed "
            "(to install ambermini do: conda install ambermini -c acellera)"
        )
    if os.path.islink(teleap):
        if os.path.isabs(os.readlink(teleap)):
            teleap = os.readlink(teleap)
        else:
            teleap = os.path.join(os.path.dirname(teleap), os.readlink(teleap))
    return teleap


def defaultAmberHome(teleap=None):
    """ Returns the default AMBERHOME as defined by the location of teLeap binary

    Parameters:
    -----------
    teleap : str
        Path to teLeap executable used to build the system for AMBER
    """
    if teleap is None:
        teleap = _findTeLeap()
    else:
        if shutil.which(teleap) is None:
            raise NameError(
                "Could not find executable: `{}` in the PATH.".format(teleap)
            )

    return os.path.normpath(os.path.join(os.path.dirname(teleap), "../"))


_defaultAmberSearchPaths = {
    "ff": join("dat", "leap", "cmd"),
    "topo": join("dat", "leap", "prep"),
    "param": join("dat", "leap", "parm"),
    "lib": join("dat", "leap", "lib"),
}


def htmdAmberHome():
    """ Returns the location of the AMBER files distributed with HTMD"""

    return os.path.abspath(os.path.join(home(shareDir=True), "builder", "amberfiles"))


def listFiles():
    """ Lists all AMBER forcefield files available in HTMD

    Example
    -------
    >>> from htmd.builder import amber
    >>> amber.listFiles()             # doctest: +ELLIPSIS
    ---- Forcefield files list: ... ----
    leaprc.constph
    leaprc.DNA.bsc1
    leaprc.DNA.OL15
    leaprc.ff14SB.redq
    ...
    """

    amberhome = defaultAmberHome()

    # Original AMBER FFs
    ffdir = join(amberhome, _defaultAmberSearchPaths["ff"])
    ffs = glob(join(ffdir, "*"))
    print("---- Forcefield files list: " + join(ffdir, "") + " ----")
    for f in sorted(ffs, key=str.lower):
        if os.path.isdir(f):
            continue
        print(f.replace(join(ffdir, ""), ""))

    oldffdir = join(amberhome, _defaultAmberSearchPaths["ff"], "oldff")
    ffs = glob(join(oldffdir, "*"))
    print("---- OLD Forcefield files list: " + join(ffdir, "") + " ----")
    for f in sorted(ffs, key=str.lower):
        print(f.replace(join(ffdir, ""), ""))

    topodir = os.path.join(amberhome, _defaultAmberSearchPaths["topo"])
    topos = glob(join(topodir, "*"))
    print("---- Topology files list: " + join(topodir, "") + " ----")
    for f in sorted(topos, key=str.lower):
        if os.path.isdir(f):
            continue
        print(f.replace(join(topodir, ""), ""))

    # FRCMOD files
    frcmoddir = os.path.join(amberhome, _defaultAmberSearchPaths["param"])
    ffs = glob(join(frcmoddir, "frcmod.*"))
    print("---- Parameter files list: " + join(frcmoddir, "") + " ----")
    for f in sorted(ffs, key=str.lower):
        print(f.replace(join(frcmoddir, ""), ""))

    htmdamberdir = htmdAmberHome()

    # Extra AMBER FFs on HTMD
    extraffs = glob(join(htmdamberdir, "*", "leaprc.*"))
    print("---- Extra forcefield files list: " + join(htmdamberdir, "") + " ----")
    for f in sorted(extraffs, key=str.lower):
        print(f.replace(join(htmdamberdir, ""), ""))

    # Extra AMBER FFs on HTMD (*.frcmod, *.in) @cuzzo87
    extratopos = glob(join(htmdamberdir, "*", "*.in"))
    print("---- Extra topology files list: " + join(htmdamberdir, "") + " ----")
    for f in sorted(extratopos, key=str.lower):
        print(f.replace(join(htmdamberdir, ""), ""))

    extraparams = glob(join(htmdamberdir, "*", "*.frcmod"))
    print("---- Extra parameter files list: " + join(htmdamberdir, "") + " ----")
    for f in sorted(extraparams, key=str.lower):
        print(f.replace(join(htmdamberdir, ""), ""))


def _locateFile(fname, ftype, teleap):
    amberhome = defaultAmberHome(teleap=teleap)
    htmdamberdir = htmdAmberHome()
    searchdir = os.path.join(amberhome, _defaultAmberSearchPaths[ftype])
    foundfile = glob(os.path.join(searchdir, fname))
    if len(foundfile) != 0:
        return foundfile[0]
    foundfile = glob(os.path.join(htmdamberdir, fname))
    if len(foundfile) != 0:
        return foundfile[0]
    logger.warning("Was not able to find {} file {}".format(ftype, fname))


def defaultFf():
    """ Returns the default leaprc forcefield files used by amber.build """
    return ["leaprc.lipid14", os.path.join("oldff", "leaprc.ff14SB"), "leaprc.gaff"]


def defaultTopo():
    """ Returns the default topology `prepi` files used by amber.build """
    return []


def defaultParam():
    """ Returns the default parameter `frcmod` files used by amber.build """

    # Common choices:
    #  * frcmod.ionsjc_tip3p:  Monovalent ion parameters for Ewald and TIP3P water from Joung & Cheatham JPCB (2008)
    #  * frcmod.ions1lm_126_tip3p + frcmod.ions234lm_126_tip3p :
    #       Li/Merz ion parameters of monovalent ions for TIP3P water model (12-6 normal usage set)
    #       Li/Merz ion parameters of divalent to tetravalent ions for TIP3P water model (12-6 normal usage set)
    #
    # See page 50 of Amber16 manual.

    return ["frcmod.ionsjc_tip3p", "frcmod.ions234lm_126_tip3p"]


def build(
    mol,
    ff=None,
    topo=None,
    param=None,
    prefix="structure",
    outdir="./build",
    caps=None,
    ionize=True,
    saltconc=0,
    saltanion=None,
    saltcation=None,
    disulfide=None,
    teleap=None,
    teleapimports=None,
    execute=True,
    atomtypes=None,
    offlibraries=None,
    gbsa=False,
    igb=2,
):
    """ Builds a system for AMBER

    Uses tleap to build a system for AMBER. Additionally it allows the user to ionize and add disulfide bridges.

    Parameters
    ----------
    mol : :class:`Molecule <moleculekit.molecule.Molecule>` object
        The Molecule object containing the system
    ff : list of str
        A list of leaprc forcefield files.
        Use :func:`amber.listFiles <htmd.builder.amber.listFiles>` to get a list of available forcefield files.
        Default: :func:`amber.defaultFf <htmd.builder.amber.defaultFf>`
    topo : list of str
        A list of topology `prepi/prep/in` files.
        Use :func:`amber.listFiles <htmd.builder.amber.listFiles>` to get a list of available topology files.
        Default: :func:`amber.defaultTopo <htmd.builder.amber.defaultTopo>`
    param : list of str
        A list of parameter `frcmod` files.
        Use :func:`amber.listFiles <htmd.builder.amber.listFiles>` to get a list of available parameter files.
        Default: :func:`amber.defaultParam <htmd.builder.amber.defaultParam>`
    prefix : str
        The prefix for the generated pdb and psf files
    outdir : str
        The path to the output directory
        Default: './build'
    caps : dict
        A dictionary with keys segids and values lists of strings describing the caps for a particular protein segment.
        e.g. caps['P'] = ['ACE', 'NME'] or caps['P'] = ['none', 'none']. Default: will apply ACE and NME caps to every
        protein segment.
    ionize : bool
        Enable or disable ionization
    saltconc : float
        Salt concentration to add to the system after neutralization.
    saltanion : {'Cl-'}
        The anion type. Please use only AMBER ion atom names.
    saltcation : {'Na+', 'K+', 'Cs+'}
        The cation type. Please use only AMBER ion atom names.
    disulfide : list of pairs of atomselection strings
        If None it will guess disulfide bonds. Otherwise provide a list pairs of atomselection strings for each pair of
        residues forming the disulfide bridge.
    teleap : str
        Path to teLeap executable used to build the system for AMBER
    teleapimports : list
        A list of paths to pass to teLeap '-I' flag, i.e. directories to be searched
        Default: determined from :func:`amber.defaultAmberHome <htmd.builder.amber.defaultAmberHome>` and
        :func:`amber.htmdAmberHome <htmd.builder.amber.htmdAmberHome>`
    execute : bool
        Disable building. Will only write out the input script needed by tleap. Does not include ionization.
    atomtypes : list of triplets
        Custom atom types defined by the user as ('type', 'element', 'hybrid') triplets
        e.g. (('C1', 'C', 'sp2'), ('CI', 'C', 'sp3')). Check `addAtomTypes` in AmberTools docs.
    offlibraries : str or list
        A path or a list of paths to OFF library files. Check `loadOFF` in AmberTools docs.
    gbsa : bool
        Modify radii for GBSA implicit water model
    igb : int
        GB model. Select: 1 for mbondi, 2 and 5 for mbondi2, 7 for bondi and 8 for mbondi3.
        Check section 4. The Generalized Born/Surface Area Model of the AMBER manual.

    Returns
    -------
    molbuilt : :class:`Molecule <moleculekit.molecule.Molecule>` object
        The built system in a Molecule object

    Example
    -------
    >>> from htmd.ui import *  # doctest: +SKIP
    >>> mol = Molecule("3PTB")
    >>> molbuilt = amber.build(mol, outdir='/tmp/build')  # doctest: +SKIP
    >>> # More complex example
    >>> disu = [['segid P and resid 157', 'segid P and resid 13'], ['segid K and resid 1', 'segid K and resid 25']]
    >>> molbuilt = amber.build(mol, outdir='/tmp/build', saltconc=0.15, disulfide=disu)  # doctest: +SKIP
    """
    # Remove pdb protein bonds as they can be regenerated by tleap. Keep non-protein bonds i.e. for ligands
    mol = mol.copy()
    _removeProteinBonds(mol)

    if teleap is None:
        teleap = _findTeLeap()
    else:
        if shutil.which(teleap) is None:
            raise NameError(
                f"Could not find executable: `{teleap}` in the PATH. Cannot build for AMBER. Please install it with `conda install ambermini -c acellera`"
            )

    if not os.path.isdir(outdir):
        os.makedirs(outdir)
    _cleanOutDir(outdir)
    if ff is None:
        ff = defaultFf()
    if topo is None:
        topo = defaultTopo()
    if param is None:
        param = defaultParam()
    if caps is None:
        caps = _defaultProteinCaps(mol)

    _missingSegID(mol)
    _checkMixedSegment(mol)

    mol = _charmmLipid2Amber(mol)

    _applyProteinCaps(mol, caps)

    f = open(os.path.join(outdir, "tleap.in"), "w")
    f.write("# tleap file generated by amber.build\n")

    # Printing out the forcefields
    for i, force in enumerate(ensurelist(ff)):
        if not os.path.isfile(force):
            force = _locateFile(force, "ff", teleap)
            if force is None:
                continue
        newname = "ff{}_{}".format(i, os.path.basename(force))
        shutil.copy(force, os.path.join(outdir, newname))
        f.write("source {}\n".format(newname))
    f.write("\n")

    if gbsa:
        gbmodels = {1: "mbondi", 2: "mbondi2", 5: "mbondi2", 7: "bondi", 8: "mbondi3"}
        f.write("set default PBradii {}\n\n".format(gbmodels[igb]))

    # Adding custom atom types
    if atomtypes is not None:
        atomtypes = ensurelist(tocheck=atomtypes[0], tomod=atomtypes)
        f.write("addAtomTypes {\n")
        for at in atomtypes:
            if len(at) != 3:
                raise RuntimeError(
                    "Atom type definitions have to be triplets. Check the AMBER documentation."
                )
            f.write('    {{ "{}" "{}" "{}" }}\n'.format(at[0], at[1], at[2]))
        f.write("}\n\n")

    # Loading OFF libraries
    if offlibraries is not None:
        offlibraries = ensurelist(offlibraries)
        for i, off in enumerate(offlibraries):
            if not os.path.isfile(off):
                raise RuntimeError(
                    "Could not find off-library in location {}".format(off)
                )
            newname = "offlib{}_{}".format(i, os.path.basename(off))
            shutil.copy(off, os.path.join(outdir, newname))
            f.write("loadoff {}\n".format(newname))

    # Loading frcmod parameters
    f.write("# Loading parameter files\n")
    for i, p in enumerate(param):
        if not os.path.isfile(p):
            p = _locateFile(p, "param", teleap)
            if p is None:
                continue
        newname = "param{}_{}".format(i, os.path.basename(p))
        shutil.copy(p, os.path.join(outdir, newname))
        f.write("loadamberparams {}\n".format(newname))
    f.write("\n")

    # Loading prepi topologies
    f.write("# Loading prepi topologies\n")
    for i, t in enumerate(topo):
        if not os.path.isfile(t):
            t = _locateFile(t, "topo", teleap)
            if t is None:
                continue
        newname = "topo{}_{}".format(i, os.path.basename(t))
        shutil.copy(t, os.path.join(outdir, newname))
        f.write("loadamberprep {}\n".format(newname))
    f.write("\n")

    f.write("# Loading the system\n")
    f.write("mol = loadpdb input.pdb\n\n")

    if np.sum(mol.atomtype != "") != 0:
        logger.debug("Writing mol2 files for input to tleap.")
        segs = np.unique(mol.segid[mol.atomtype != ""])
        combstr = "mol = combine {mol"
        for s in segs:
            name = "segment{}".format(s)
            mol2name = os.path.join(outdir, "{}.mol2".format(name))
            mol.write(mol2name, (mol.atomtype != "") & (mol.segid == s))
            if not os.path.isfile(mol2name):
                raise NameError(
                    "Could not write a mol2 file out of the given Molecule."
                )
            f.write("# Loading the rest of the system\n")
            f.write("{} = loadmol2 {}.mol2\n\n".format(name, name))
            combstr += " {}".format(name)
        combstr += "}\n\n"
        f.write(combstr)

    # Write patches for disulfide bonds (only after ionizing)
    if not ionize:
        # TODO: Remove this once we deprecate the class
        from htmd.builder.builder import DisulfideBridge
        from moleculekit.molecule import UniqueResidueID

        if (
            disulfide is not None
            and len(disulfide) != 0
            and isinstance(disulfide[0], DisulfideBridge)
        ):
            newdisu = []
            for d in disulfide:
                r1 = UniqueResidueID.fromMolecule(
                    mol, "resid {} and segname {}".format(d.resid1, d.segid1)
                )
                r2 = UniqueResidueID.fromMolecule(
                    mol, "resid {} and segname {}".format(d.resid2, d.segid2)
                )
                newdisu.append([r1, r2])
            disulfide = newdisu
        # TODO: Remove up to here ----------------------

        if (
            disulfide is not None
            and len(disulfide) != 0
            and isinstance(disulfide[0][0], str)
        ):
            disulfide = convertDisulfide(mol, disulfide)

        if disulfide is None:
            logger.info("Detecting disulfide bonds.")
            disulfide = detectDisulfideBonds(mol)

        # Fix structure to match the disulfide patching
        if len(disulfide) != 0:
            torem = np.zeros(mol.numAtoms, dtype=bool)
            f.write("# Adding disulfide bonds\n")
            for d in disulfide:
                # Rename the residues to CYX if there is a disulfide bond
                atoms1 = d[0].selectAtoms(mol, indexes=False)
                atoms2 = d[1].selectAtoms(mol, indexes=False)
                mol.resname[atoms1] = "CYX"
                mol.resname[atoms2] = "CYX"
                # Remove (eventual) HG hydrogens on these CYS (from proteinPrepare)
                torem |= (atoms1 & (mol.name == "HG")) | (atoms2 & (mol.name == "HG"))
                # Convert to stupid amber residue numbering
                uqseqid = (
                    sequenceID((mol.resid, mol.insertion, mol.segid)) + mol.resid[0]
                )
                uqres1 = int(np.unique(uqseqid[atoms1]))
                uqres2 = int(np.unique(uqseqid[atoms2]))
                f.write("bond mol.{}.SG mol.{}.SG\n".format(uqres1, uqres2))
            f.write("\n")
            mol.remove(torem, _logger=False)

    # Calculate the bounding box and store it in the CRD file
    f.write('setBox mol "vdw"\n\n')

    f.write("# Writing out the results\n")
    f.write("saveamberparm mol " + prefix + ".prmtop " + prefix + ".crd\n")
    f.write("quit")
    f.close()

    # Printing and loading the PDB file. AMBER can work with a single PDB file if the segments are separate by TER
    logger.debug("Writing PDB file for input to tleap.")
    pdbname = os.path.join(outdir, "input.pdb")

    # mol2 files have atomtype, here we only write parts not coming from mol2
    # We need to write the input.pdb at the end since we modify the resname for disulfide bridges in mol
    mol.write(pdbname, mol.atomtype == "")
    if not os.path.isfile(pdbname):
        raise NameError("Could not write a PDB file out of the given Molecule.")

    molbuilt = None
    if execute:
        if not teleapimports:
            teleapimports = []
            # Source default Amber (i.e. the same paths tleap imports)
            amberhome = defaultAmberHome(teleap=teleap)
            teleapimports += [
                os.path.join(amberhome, s) for s in _defaultAmberSearchPaths.values()
            ]
            if len(teleapimports) == 0:
                raise RuntimeWarning(
                    "No default Amber force-field found. Check teLeap location: {}".format(
                        teleap
                    )
                )
            # Source HTMD Amber paths that contain ffs
            htmdamberdir = htmdAmberHome()
            teleapimports += [
                os.path.join(htmdamberdir, os.path.dirname(f))
                for f in ff
                if os.path.isfile(os.path.join(htmdamberdir, f))
            ]
            if len(teleapimports) == 0:
                raise RuntimeError(
                    "No default Amber force-field imports found. Check "
                    "`htmd.builder.amber.defaultAmberHome()` and `htmd.builder.amber.htmdAmberHome()`"
                )
        # Set import flags for teLeap
        teleapimportflags = []
        for p in teleapimports:
            teleapimportflags.append("-I")
            teleapimportflags.append(str(p))
        logpath = os.path.abspath(os.path.join(outdir, "log.txt"))
        logger.info("Starting the build.")
        currdir = os.getcwd()
        os.chdir(outdir)
        f = open(logpath, "w")
        try:
            cmd = [teleap, "-f", "./tleap.in"]
            cmd[1:1] = teleapimportflags
            logger.debug(cmd)
            call(cmd, stdout=f)
        except:
            raise NameError("teLeap failed at execution")
        f.close()
        errors = _logParser(logpath)
        os.chdir(currdir)
        if errors:
            raise BuildError(
                errors
                + [
                    "Check {} for further information on errors in building.".format(
                        logpath
                    )
                ]
            )
        logger.info("Finished building.")

        if (
            os.path.exists(os.path.join(outdir, "structure.crd"))
            and os.path.getsize(os.path.join(outdir, "structure.crd")) != 0
            and os.path.getsize(os.path.join(outdir, "structure.prmtop")) != 0
        ):
            try:
                molbuilt = Molecule(os.path.join(outdir, "structure.prmtop"))
                molbuilt.read(os.path.join(outdir, "structure.crd"))
            except Exception as e:
                raise RuntimeError(
                    "Failed at reading structure.prmtop/structure.crd due to error: {}".format(
                        e
                    )
                )
        else:
            raise BuildError(
                "No structure pdb/prmtop file was generated. Check {} for errors in building.".format(
                    logpath
                )
            )

        if ionize:
            shutil.move(
                os.path.join(outdir, "structure.crd"),
                os.path.join(outdir, "structure.noions.crd"),
            )
            shutil.move(
                os.path.join(outdir, "structure.prmtop"),
                os.path.join(outdir, "structure.noions.prmtop"),
            )
            totalcharge = np.sum(molbuilt.charge)
            nwater = np.sum(molbuilt.atomselect("water and noh"))
            anion, cation, anionatom, cationatom, nanion, ncation = ionizef(
                totalcharge,
                nwater,
                saltconc=saltconc,
                anion=saltanion,
                cation=saltcation,
            )
            newmol = ionizePlace(
                mol, anion, cation, anionatom, cationatom, nanion, ncation
            )
            # Redo the whole build but now with ions included
            return build(
                newmol,
                ff=ff,
                topo=topo,
                param=param,
                prefix=prefix,
                outdir=outdir,
                caps={},
                ionize=False,
                execute=execute,
                saltconc=saltconc,
                disulfide=disulfide,
                teleap=teleap,
                atomtypes=atomtypes,
                offlibraries=offlibraries,
            )

    tmpbonds = molbuilt.bonds
    molbuilt.bonds = []  # Removing the bonds to speed up writing
    molbuilt.write(os.path.join(outdir, "structure.pdb"))
    molbuilt.bonds = tmpbonds  # Restoring the bonds
    detectCisPeptideBonds(molbuilt)  # Warn in case of cis bonds
    return molbuilt


def _applyProteinCaps(mol, caps):

    # AMBER capping
    # =============
    # This is the (horrible) way of adding caps in tleap:
    # For now, this is hardwired for ACE and NME
    # 1. Change one of the hydrogens of the N terminal (H[T]?[123]) to the ACE C atom, giving it a new resid
    # 1a. If no hydrogen present, create the ACE C atom.
    # 2. Change one of the oxygens of the C terminal ({O,OT1,OXT}) to the NME N atom, giving it a new resid
    # 2a. If no oxygen present, create the NME N atom.
    # 3. Reorder to put the new atoms first and last
    # 4. Remove the lingering hydrogens of the N terminal and oxygens of the C terminal.

    # Define the atoms to be replaced (0 and 1 corresponds to N- and C-terminal caps)
    terminalatoms = {
        "ACE": ["H1", "H2", "H3", "HT1", "HT2", "HT3"],
        "NME": ["OXT", "OT1", "O"],
    }  # XPLOR names for H[123] and OXT are HT[123]
    # and OT1, respectively.
    capresname = ["ACE", "NME"]
    capatomtype = ["C", "N"]

    # For each caps definition
    for seg in caps:
        prot = mol.atomselect(
            "protein"
        )  # Can't move this out since we remove atoms in this loop
        # Get the segment
        segment = np.where(mol.segid == seg)[0]
        # Test segment
        if len(segment) == 0:
            raise RuntimeError("There is no segment {} in the molecule.".format(seg))
        if not np.any(prot & (mol.segid == seg)):
            raise RuntimeError(
                "Segment {} is not protein. Capping for non-protein segments is not supported.".format(
                    seg
                )
            )
        # For each cap
        passed = False
        for i, cap in enumerate(caps[seg]):
            if cap is None or (isinstance(cap, str) and cap == "none"):
                continue
            # Get info on segment and its terminals
            segidm = mol.segid == seg  # Mask for segid
            segididx = np.where(segidm)[0]
            resids = mol.resid[segididx]
            terminalids = [segididx[0], segididx[-1]]
            terminalresids = [resids[0], resids[-1]]
            residm = mol.resid == terminalresids[i]  # Mask for resid

            if not passed:
                orig_terminalresids = terminalresids
                passed = True

            if cap is None or cap == "":  # In case there is no cap defined
                logger.warning(
                    "No cap provided for resid {} on segment {}. Did not apply it.".format(
                        terminalresids[i], seg
                    )
                )
                continue
            elif cap not in capresname:  # If it is defined, test if supported
                raise RuntimeError(
                    "In segment {}, the {} cap is not supported. Try using {} instead.".format(
                        seg, cap, capresname
                    )
                )

            # Test if cap is already applied
            testcap = np.where(segidm & residm & (mol.resname == cap))[0]
            if len(testcap) != 0:
                logger.warning(
                    "Cap {} already exists on segment {}. Did not re-apply it.".format(
                        cap, seg
                    )
                )
                continue

            # Test if the atom to change exists
            termatomsids = np.zeros(residm.shape, dtype=bool)
            for atm in terminalatoms[cap]:
                termatomsids |= mol.name == atm
            termatomsids = np.where(termatomsids & segidm & residm)[0]

            if len(termatomsids) == 0:
                # Create new atom
                termcaid = np.where(segidm & residm & (mol.name == "CA"))[0]
                termcenterid = np.where(
                    segidm & residm & (mol.name == capatomtype[1 - i])
                )[0]
                atom = Molecule()
                atom.empty(1)
                atom.record = np.array(["ATOM"], dtype=Molecule._dtypes["record"])
                atom.name = np.array([capatomtype[i]], dtype=Molecule._dtypes["name"])
                atom.resid = np.array(
                    [terminalresids[i] - 1 + 2 * i], dtype=Molecule._dtypes["resid"]
                )
                atom.resname = np.array([cap], dtype=Molecule._dtypes["resname"])
                atom.segid = np.array([seg], dtype=Molecule._dtypes["segid"])
                atom.element = np.array(
                    [capatomtype[i]], dtype=Molecule._dtypes["element"]
                )
                atom.chain = np.array(
                    [np.unique(mol.chain[segidm])], dtype=Molecule._dtypes["chain"]
                )  # TODO: Assumption of single chain in a segment might be wrong
                atom.coords = mol.coords[termcenterid] + 0.33 * np.subtract(
                    mol.coords[termcenterid], mol.coords[termcaid]
                )
                mol.insert(atom, terminalids[i])
                # logger.info('In segment {}, resid {} had none of these atoms: {}. Capping was performed by creating '
                #             'a new atom for cap construction by tleap.'.format(seg, terminalresids[i],
                #                                                                ' '.join(terminalatoms[cap])))

            else:
                # Select atom to change, do changes to cap, and change resid
                newatom = np.max(termatomsids)
                mol.set("resname", cap, sel=newatom)
                mol.set("name", capatomtype[i], sel=newatom)
                mol.set("element", capatomtype[i], sel=newatom)
                mol.set(
                    "resid", terminalresids[i] - 1 + 2 * i, sel=newatom
                )  # if i=0 => resid-1; i=1 => resid+1

                # Reorder
                neworder = np.arange(mol.numAtoms)
                neworder[newatom] = terminalids[i]
                neworder[terminalids[i]] = newatom
                _reorderMol(mol, neworder)

        # For each cap
        for i, cap in enumerate(caps[seg]):
            if cap is None or (isinstance(cap, str) and cap == "none"):
                continue
            # Remove lingering hydrogens or oxygens in terminals
            mol.remove(
                'segid {} and resid "{}" and name {}'.format(
                    seg, orig_terminalresids[i], " ".join(terminalatoms[cap])
                ),
                _logger=False,
            )

    # Remove terminal hydrogens regardless of caps or no caps. tleap confuses itself when it makes residues into terminals.
    # For example HID has an atom H which in NHID should become H[123] and crashes tleap.
    for seg in np.unique(mol.get("segid", "protein")):
        segidm = mol.segid == seg  # Mask for segid
        segididx = np.where(segidm)[0]
        resids = mol.resid[segididx]
        mol.remove(
            '(resid "{}" "{}") and segid {} and hydrogen'.format(
                resids[0], resids[-1], seg
            ),
            _logger=False,
        )


def _removeProteinBonds(mol):
    segs = np.unique(
        mol.segid[mol.atomtype != ""]
    )  # Keeping bonds related to mol2 files
    mol.deleteBonds(~np.in1d(mol.segid, segs))


def _defaultProteinCaps(mol):
    # Defines ACE and NME (neutral terminals) as default for protein segments
    # Of course, this might not be ideal for proteins that require charged terminals

    segsProt = np.unique(mol.get("segid", sel="protein"))
    caps = dict()
    for s in segsProt:
        if len(np.unique(mol.resid[mol.segid == s])) < 10:
            logger.warning(
                "Segment {} consists of a peptide with less than 10 residues. It will not be capped by "
                "default. If you want to cap it use the caps argument of amber.build to manually define caps"
                "for all segments.".format(s)
            )
            continue
        caps[s] = ["ACE", "NME"]
    return caps


def _cleanOutDir(outdir):
    from glob import glob

    files = glob(os.path.join(outdir, "structure.*"))
    files += glob(os.path.join(outdir, "log.*"))
    files += glob(os.path.join(outdir, "*.log"))
    for f in files:
        os.remove(f)


def _charmmLipid2Amber(mol):
    """ Convert a CHARMM lipid membrane to AMBER format

    Parameters
    ----------
    mol : :class:`Molecule <moleculekit.molecule.Molecule>` object
        The Molecule object containing the membrane

    Returns
    -------
    newmol : :class:`Molecule <moleculekit.molecule.Molecule>` object
        A new Molecule object with the membrane converted to AMBER
    """

    resdict = _readcsvdict(
        os.path.join(home(shareDir=True), "builder", "charmmlipid2amber.csv")
    )

    natoms = mol.numAtoms
    neworder = np.array(
        list(range(natoms))
    )  # After renaming the atoms and residues I have to reorder them

    begs = np.zeros(natoms, dtype=bool)
    fins = np.zeros(natoms, dtype=bool)
    begters = np.zeros(natoms, dtype=bool)
    finters = np.zeros(natoms, dtype=bool)

    # Iterate over the translation dictionary
    mol = mol.copy()
    incrresids = sequenceID((mol.resid, mol.insertion, mol.segid))

    for res in resdict.keys():
        molresidx = mol.resname == res
        if not np.any(molresidx):
            continue
        names = (
            mol.name.copy()
        )  # Need to make a copy or I accidentally double-modify atoms

        atommap = resdict[res]
        for atom in atommap.keys():
            rule = atommap[atom]

            molatomidx = np.zeros(len(names), dtype=bool)
            molatomidx[molresidx] = names[molresidx] == atom

            mol.set("resname", rule.replaceresname, sel=molatomidx)
            mol.set("name", rule.replaceatom, sel=molatomidx)
            neworder[molatomidx] = rule.order

            if rule.order == 0:  # First atom (with or without ters)
                begs[molatomidx] = True
            if rule.order == rule.natoms - 1:  # Last atom (with or without ters)
                fins[molatomidx] = True
            if rule.order == 0 and rule.ter:  # First atom with ter
                begters[molatomidx] = True
            if rule.order == rule.natoms - 1 and rule.ter:  # Last atom with ter
                finters[molatomidx] = True

    uqresids = np.unique(incrresids[begs])
    residuebegs = np.ones(len(uqresids), dtype=int) * -1
    residuefins = np.ones(len(uqresids), dtype=int) * -1
    for i in range(len(uqresids)):
        residuebegs[i] = np.where(incrresids == uqresids[i])[0][0]
        residuefins[i] = np.where(incrresids == uqresids[i])[0][-1]
    for i in range(len(residuebegs)):
        beg = residuebegs[i]
        fin = residuefins[i] + 1
        neworder[beg:fin] = neworder[beg:fin] + beg
    idx = np.argsort(neworder)

    _reorderMol(mol, idx)

    begters = np.where(begters[idx])[0]  # Sort the begs and ters
    finters = np.where(finters[idx])[0]

    # if len(begters) > 999:
    #    raise NameError('More than 999 lipids. Cannot define separate segments for all of them.')

    for i in range(len(begters)):
        map = np.zeros(len(mol.resid), dtype=bool)
        map[begters[i] : finters[i] + 1] = True
        mol.set("resid", sequenceID(mol.get("resname", sel=map)), sel=map)
        mol.set("segid", "L{}".format(i % 2), sel=map)

    return mol


def _reorderMol(mol, order):
    for k in mol._atom_and_coord_fields:
        if mol.__dict__[k] is not None and np.size(mol.__dict__[k]) != 0:
            if k == "coords":
                mol.__dict__[k] = mol.__dict__[k][order, :, :]
            else:
                mol.__dict__[k] = mol.__dict__[k][order]


def _readcsvdict(filename):
    import csv
    from collections import namedtuple

    if os.path.isfile(filename):
        csvfile = open(filename, "r")
    else:
        raise NameError("File " + filename + " does not exist")

    resdict = dict()

    Rule = namedtuple(
        "Rule", ["replaceresname", "replaceatom", "order", "natoms", "ter"]
    )

    # Skip header line of csv file. Line 2 contains dictionary keys:
    csvfile.readline()
    csvreader = csv.DictReader(csvfile)
    for line in csvreader:
        searchres = line["search"].split()[1]
        searchatm = line["search"].split()[0]
        if searchres not in resdict:
            resdict[searchres] = dict()
        resdict[searchres][searchatm] = Rule(
            line["replace"].split()[1],
            line["replace"].split()[0],
            int(line["order"]),
            int(line["num_atom"]),
            line["TER"] == "True",
        )
    csvfile.close()

    return resdict


def _logParser(fname):
    import re

    unknownres_regex = re.compile("Unknown residue:\s+(\w+)")
    missingparam_regex = re.compile(
        "For atom: (.*) Could not find vdW \(or other\) parameters for type:\s+(\w+)"
    )
    missingtorsion_regex = re.compile("No torsion terms for\s+(.*)$")
    missingbond_regex = re.compile("Could not find bond parameter for:\s+(.*)$")
    missingangle_regex = re.compile("Could not find angle parameter:\s+(.*)$")
    missingatomtype_regex = re.compile("FATAL:\s+Atom (.*) does not have a type")

    unknownres = []
    missingparam = []
    missingtorsion = []
    missingbond = []
    missingangle = []
    missingatomtype = []
    with open(fname, "r") as f:
        for line in f:
            if unknownres_regex.search(line):
                unknownres.append(unknownres_regex.findall(line)[0])
            if missingparam_regex.search(line):
                missingparam.append(missingparam_regex.findall(line)[0])
            if missingtorsion_regex.search(line):
                missingtorsion.append(missingtorsion_regex.findall(line)[0])
            if missingbond_regex.search(line):
                missingbond.append(missingbond_regex.findall(line)[0])
            if missingangle_regex.search(line):
                missingangle.append(missingangle_regex.findall(line)[0])
            if missingatomtype_regex.search(line):
                missingatomtype.append(missingatomtype_regex.findall(line)[0])

    errors = []
    if len(unknownres):
        errors.append(
            UnknownResidueError(
                "Unknown residue(s) {} found in the input structure. "
                "You are either missing a topology definition for the residue or you need to "
                "rename it to the correct residue name".format(np.unique(unknownres))
            )
        )
    if len(missingparam):
        errors.append(
            MissingParameterError(
                "Missing parameters for atom {} and type {}".format(
                    missingparam, missingparam
                )
            )
        )
    if len(missingtorsion):
        errors.append(
            MissingTorsionError("Missing torsion terms for {}".format(missingtorsion))
        )
    if len(missingbond):
        errors.append(
            MissingBondError("Missing bond parameters for {}".format(missingbond))
        )
    if len(missingangle):
        errors.append(
            MissingAngleError("Missing angle parameters for {}".format(missingangle))
        )
    if len(missingatomtype):
        errors.append(
            MissingAtomTypeError("Missing atom type for {}".format(missingatomtype))
        )

    return errors


class _TestAmberBuild(unittest.TestCase):
    currentResult = None  # holds last result object passed to run method

    def setUp(self):
        from htmd.util import tempname

        self.testDir = os.environ.get("TESTDIR", tempname())
        print("Running tests in {}".format(self.testDir))

    def run(self, result=None):
        self.currentResult = result  # remember result for use in tearDown
        unittest.TestCase.run(self, result)  # call superclass run method

    def tearDown(self):
        pass
        # if self.currentResult.wasSuccessful():
        #     shutil.rmtree(self.testDir)

    @staticmethod
    def _compareResultFolders(
        compare, tmpdir, pid, ignore_ftypes=(".log", ".txt", ".frcmod")
    ):
        import filecmp

        def _cutfirstline(infile, outfile):
            # Cut out the first line of prmtop which has a build date in it
            with open(infile, "r") as fin:
                data = fin.read().splitlines(True)
            with open(outfile, "w") as fout:
                fout.writelines(data[1:])

        files = []
        deletefiles = []
        for f in glob(join(compare, "*")):
            fname = os.path.basename(f)
            if os.path.splitext(f)[1] in ignore_ftypes:
                continue
            if f.endswith("prmtop"):
                _cutfirstline(f, join(compare, fname + ".mod"))
                _cutfirstline(join(tmpdir, fname), os.path.join(tmpdir, fname + ".mod"))
                files.append(os.path.basename(f) + ".mod")
                deletefiles.append(join(compare, fname + ".mod"))
            else:
                files.append(os.path.basename(f))

        match, mismatch, error = filecmp.cmpfiles(tmpdir, compare, files, shallow=False)
        if len(mismatch) != 0 or len(error) != 0 or len(match) != len(files):
            raise RuntimeError(
                "Different results produced by amber.build for test {} between {} and {} in files {}.".format(
                    pid, compare, tmpdir, mismatch
                )
            )

        for f in deletefiles:
            os.remove(f)

    def testWithProteinPrepare(self):
        from moleculekit.tools.preparation import proteinPrepare
        from htmd.builder.solvate import solvate

        # Test with proteinPrepare
        pdbids = ["3PTB"]
        # pdbids = ['3PTB', '1A25', '1GZM']  # '1U5U' out because it has AR0 (no parameters)
        for pid in pdbids:
            np.random.seed(1)
            mol = Molecule(pid)
            mol.filter("protein")
            mol = proteinPrepare(mol)
            mol.filter("protein")  # Fix for bad proteinPrepare hydrogen placing
            smol = solvate(mol)
            ffs = defaultFf()
            tmpdir = os.path.join(self.testDir, "withProtPrep", pid)
            _ = build(smol, ff=ffs, outdir=tmpdir)

            refdir = home(dataDir=join("test-amber-build", "pp", pid))
            _TestAmberBuild._compareResultFolders(refdir, tmpdir, pid)

    def testWithoutProteinPrepare(self):
        from htmd.builder.solvate import solvate

        # Test without proteinPrepare
        pdbids = ["3PTB"]
        # pdbids = ['3PTB', '1A25', '1GZM', '1U5U']
        for pid in pdbids:
            np.random.seed(1)
            mol = Molecule(pid)
            mol.filter("protein")
            smol = solvate(mol)
            ffs = defaultFf()
            tmpdir = os.path.join(self.testDir, "withoutProtPrep", pid)
            _ = build(smol, ff=ffs, outdir=tmpdir)

            refdir = home(dataDir=join("test-amber-build", "nopp", pid))
            _TestAmberBuild._compareResultFolders(refdir, tmpdir, pid)

    def testProteinLigand(self):
        from htmd.builder.solvate import solvate

        # Test protein ligand building with parametrized ligand
        refdir = home(dataDir=join("test-amber-build", "protLig"))
        tmpdir = os.path.join(self.testDir, "protLig")
        os.makedirs(tmpdir)

        mol = Molecule(join(refdir, "3ptb_mod.pdb"))
        lig = Molecule(join(refdir, "benzamidine.mol2"))
        lig.segid[:] = "L"

        # params =
        newmol = Molecule()
        newmol.append(lig)
        newmol.append(mol)
        smol = solvate(newmol)

        params = defaultParam() + [
            join(refdir, "benzamidine.frcmod"),
        ]

        _ = build(smol, outdir=tmpdir, param=params, ionize=False)

        refdir = home(dataDir=join("test-amber-build", "protLig", "results"))
        _TestAmberBuild._compareResultFolders(refdir, tmpdir, "3PTB")

        # # Test protein-ligand building
        # folder = home(dataDir='building-protein-ligand')
        # prot = Molecule(os.path.join(folder, 'trypsin.pdb'))
        # prot.filter('protein')
        # prot = autoSegment2(prot)
        # prot = proteinPrepare(prot)
        # prot1 = prot
        # prot2 = prot.copy()
        # lig1 = Molecule(os.path.join(folder, 'benzamidine.mol2'))
        # lig1.set('segid', 'L')
        # lig2 = Molecule(os.path.join(folder, 'benzamidine.pdb'))
        # lig2.set('segid', 'L')
        # prot1.append(lig1)
        # prot2.append(lig2)
        # smol1 = solvate(prot1)
        # smol2 = solvate(prot2)
        # tmpdir1 = tempname()
        # tmpdir2 = tempname()
        # np.random.seed(1)
        # params = amber.defaultParam() + [os.path.join(folder, 'benzamidine.frcmod')]
        # bmol1 = amber.build(smol1, param=params, outdir=tmpdir1)
        # np.random.seed(1)
        # params = amber.defaultParam() + [os.path.join(folder, 'benzamidineprepi.frcmod')]
        # topos = amber.defaultTopo() + [os.path.join(folder, 'benzamidine.prepi')]
        # bmol2 = amber.build(smol2, topo=topos, param=params, outdir=tmpdir2)
        # _compareResultFolders(tmpdir1, tmpdir2, 'ben-tryp')
        # shutil.rmtree(tmpdir1)
        # shutil.rmtree(tmpdir2)

    def test_customDisulfideBonds(self):
        from htmd.builder.solvate import solvate

        # Test without proteinPrepare
        pdbids = [
            "1GZM",
        ]
        for pid in pdbids:
            np.random.seed(1)
            mol = Molecule(pid)
            mol.filter("protein")
            smol = solvate(mol)
            ffs = defaultFf()
            disu = [
                ["segid 1 and resid 110", "segid 1 and resid 187"],
                ["segid 0 and resid 110", "segid 0 and resid 187"],
            ]
            tmpdir = os.path.join(self.testDir, "withoutProtPrep", pid)
            _ = build(smol, ff=ffs, outdir=tmpdir, disulfide=disu)

            refdir = home(dataDir=join("test-amber-build", "nopp", pid))
            _TestAmberBuild._compareResultFolders(refdir, tmpdir, pid)

            np.random.seed(1)
            tmpdir = os.path.join(self.testDir, "withoutProtPrep", pid)
            _ = build(smol, ff=ffs, outdir=tmpdir)

            refdir = home(dataDir=join("test-amber-build", "nopp", pid))
            _TestAmberBuild._compareResultFolders(refdir, tmpdir, pid)

    def test_customTeLeapImports(self):
        from htmd.builder.solvate import solvate

        # Test without proteinPrepare
        pdbids = ["3PTB"]
        # pdbids = ['3PTB', '1A25', '1GZM', '1U5U']
        for pid in pdbids:
            np.random.seed(1)
            mol = Molecule(pid)
            mol.filter("protein")
            smol = solvate(mol)
            ffs = defaultFf()
            tmpdir = os.path.join(self.testDir, "withoutProtPrep", pid)

            amberhome = defaultAmberHome()
            teleapimports = []
            teleapimports.append(
                os.path.join(amberhome, _defaultAmberSearchPaths["lib"])
            )
            teleapimports.append(
                os.path.join(amberhome, _defaultAmberSearchPaths["param"])
            )
            _ = build(smol, ff=ffs, outdir=tmpdir, teleapimports=teleapimports)

            refdir = home(dataDir=join("test-amber-build", "nopp", pid))
            _TestAmberBuild._compareResultFolders(refdir, tmpdir, pid)


if __name__ == "__main__":
    import doctest
    import unittest

    failure_count, _ = doctest.testmod()
    unittest.main(verbosity=2)
    # if failure_count != 0:
    #     raise Exception('Doctests failed')




# (c) 2015-2018 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
from simtk import openmm
from simtk import unit
from simtk.openmm import app
import numpy as np


def _halton_sequence(p, n):
    """
    Halton deterministic sequence on [0,1].
    Parameters
    ----------
    p : int
       Prime number for sequence.
    n : int
       Sequence length to generate.
    Returns
    -------
    u : numpy.array of double
       Sequence on [0,1].
    Notes
    -----
    Code source: http://blue.math.buffalo.edu/sauer2py/
    More info: http://en.wikipedia.org/wiki/Halton_sequence
    Examples
    --------
    Generate some sequences with different prime number bases.
    >>> x = _halton_sequence(2,100)
    >>> y = _halton_sequence(3,100)
    >>> z = _halton_sequence(5,100)
    """
    eps = np.finfo(np.double).eps
    # largest number of digits (adding one for halton_sequence(2,64) corner case)
    b = np.zeros(int(np.ceil(np.log(n) / np.log(p))) + 1)
    u = np.empty(n)
    for j in range(n):
        i = 0
        b[0] += 1                       # add one to current integer
        while b[i] > p - 1 + eps:           # this loop does carrying in base p
            b[i] = 0
            i = i + 1
            b[i] += 1
        u[j] = 0
        for k in range(len(b)):         # add up reversed digits
            u[j] += b[k] * p**-(k + 1)
    return u


def _subrandom_particle_positions(nparticles, box_vectors, ndim, method='halton'):
    """Generate a deterministic list of subrandom particle positions."""
    # Create positions array.
    positions = np.zeros([nparticles, 3], np.float32)

    # Fill in each dimension.
    primes = [2, 3, 5]  # prime bases for Halton sequence
    for dim in range(ndim):
        x = _halton_sequence(primes[dim], nparticles)
        l = box_vectors[dim][dim]
        positions[:, dim] = x - 0.5

    np.random.shuffle(positions)
    return unit.Quantity(positions * l / l.unit, l.unit)


def distributeLipids(boxsize,
                     resnames,
                     sigmas,
                     cutoff,
                     mass=39.9 * unit.amu,  # argon
                     epsilon=0.238 * unit.kilocalories_per_mole,  # argon,
                     switch_width=3.4 * unit.angstrom,  # argon
                     ):
        nparticles = len(resnames)
                
        # Determine Lennard-Jones cutoff.
        cutoff = cutoff * unit.angstrom

        cutoff_type = openmm.NonbondedForce.CutoffPeriodic

        # Create an empty system object.
        system = openmm.System()

        # Periodic box vectors.
        a = unit.Quantity((boxsize[0] * unit.angstrom, 0 * unit.angstrom, 0 * unit.angstrom))
        b = unit.Quantity((0 * unit.angstrom, boxsize[1] * unit.angstrom, 0 * unit.angstrom))
        c = unit.Quantity((0 * unit.angstrom, 0 * unit.angstrom, boxsize[2] * unit.angstrom))
        system.setDefaultPeriodicBoxVectors(a, b, c)

        # Set up periodic nonbonded interactions with a cutoff.
        nb = openmm.NonbondedForce()
        nb.setNonbondedMethod(cutoff_type)
        nb.setCutoffDistance(cutoff)
        nb.setUseDispersionCorrection(True)

        nb.setUseSwitchingFunction(False)
        if (switch_width is not None):
            nb.setUseSwitchingFunction(True)
            nb.setSwitchingDistance(cutoff - switch_width)

        for s in sigmas:
            system.addParticle(mass)
            nb.addParticle(0.0 * unit.elementary_charge, s * unit.angstrom, epsilon)

        positions = _subrandom_particle_positions(nparticles, system.getDefaultPeriodicBoxVectors(), 2)

        # Add the nonbonded force.
        system.addForce(nb)

        # Add a restraining potential to keep atoms in z=0
        energy_expression = 'k * (z^2)'
        force = openmm.CustomExternalForce(energy_expression)
        force.addGlobalParameter('k', 10)
        for particle_index in range(nparticles):
            force.addParticle(particle_index, [])
        system.addForce(force)

        # Create topology.
        topology = app.Topology()
        chain = topology.addChain()
        elems = ['Ar', 'Cl', 'Na']
        _, idx = np.unique(resnames, return_inverse=True)
        for i in idx:
            element = app.Element.getBySymbol(elems[i])
            residue = topology.addResidue(elems[i], chain)
            topology.addAtom(elems[i], element, residue)

        topology.setUnitCellDimensions(unit.Quantity(boxsize, unit.angstrom)) 
            
        # Simulate it
        from simtk.openmm import VerletIntegrator
        from simtk.openmm.app import Simulation
        from simtk.unit import picoseconds, picosecond, angstrom
        nsteps = 10000

        integrator = VerletIntegrator(0.002 * picoseconds)
        simulation = Simulation(topology, system, integrator)
        simulation.context.setPositions(positions)
        simulation.minimizeEnergy()
        # simulation.reporters.append(DCDReporter('output.dcd', 1))
        # simulation.reporters.append(StateDataReporter(stdout, 1000, potentialEnergy=True, totalEnergy=True, step=True, separator='   '))
        simulation.step(nsteps)

        state = simulation.context.getState(getPositions=True, enforcePeriodicBox=True)
        allfinalpos = state.getPositions(asNumpy=True).value_in_unit(angstrom)

        return allfinalpos


if __name__ == '__main__':
    pass



# (c) 2015-2018 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#

# Author: Juan Eiros Zamora (je714@ic.ac.uk) & Gil Ferreira Hoben (gf712@ic.ac.uk) Organisation: Imperial College London
# Modified by: Stefan Doerr

from moleculekit.molecule import Molecule
from htmd.apps.pmemd import Pmemd
from htmd.protocols.oldprotocolinterface import ProtocolInterface, TYPE_INT, TYPE_FLOAT, RANGE_0POS, RANGE_POS, RANGE_ANY
import os
import numpy as np
import logging
import shutil
import re
from htmd.decorators import _Deprecated
logger = logging.getLogger(__name__)


@_Deprecated('1.13.6')
class Production(ProtocolInterface):

    """ Production protocol for globular and membrane proteins
        Example
        -------
        >>> from htmd.protocols.equilibration_v1 import Equilibration
        >>> md = Equilibration()
        >>> md.numsteps = 10000000
        >>> md.temperature = 300
        >>> md.useconstantratio = True  # only for membrane sims
        >>> # this is only needed for setting the flatbottom potential, otherwise remove it
        >>> md.reference = 'protein and resid 293'
        >>> md.selection = 'segname L and noh'
        >>> md.box = [-25, 25, -25, 25, 43, 45]
        >>> md.k = 5
        >>> md.write('./build','./equil')
    """

    def __init__(self):
        super().__init__()
        self._cmdObject('amber', ':class:`MDEngine <htmd.apps.app.App>` object', 'MD engine', None, Pmemd)

        self.amber = Pmemd()
        self.amber.imin = 0
        self.amber.ntx = 5
        self.amber.irest = 1
        self.amber.ntc = 2
        self.amber.ntf = 2
        self.amber.nstlim = 12500000
        self.amber.ntt = 3
        self.amber.gamma_ln = 1.0
        self.amber.temp0 = 300.0
        self.amber.ntpr = 5000
        self.amber.ntwr = 50000
        self.amber.ntwx = 5000
        self.amber.dt = 0.004
        self.amber.ig = -1
        self.amber.ntb = 2
        self.amber.ntp = 1
        self.amber.cut = 8.0
        self.amber.barostat = 2
        self.amber.ioutfm = 1
        self.amber.ntxo = 2
        self.amber.outputnc = 'Production.nc'
        self.amber.FORTRAN = """ Production
 &cntrl
   imin=IMIN, ntx=NTX, irest=IREST,
   ntc=NTC, ntf=NTF,
   nstlim=NSTLIM, ntt=NTT, gamma_ln=GAMMA_LN,
   temp0=TEMP0,
   ntpr=NTPR, ntwr=NTWR, ntwx=NTWX,
   dt=DT, ig=IG,
   ntb=NTB, ntp=NTP, cut=CUT, barostat=BAROSTAT, ioutfm=IOUTFM, ntxo=NTXO,
 / \n"""
        # self.amber.FORTRAN = ''' HEATING\n &cntrl\n'''
        self.amber.bash = '''ENGINE -O -i INPUT -o OUTPUT -p TOPOLOGY -c RESTART -x TRAJOUT -r RESTOUT'''  # FIXME: add -ref REFERENCE

    def _findFiles(self, inputdir):
        # Tries to find default files if the given don't exist
        defaults = {'coordinates': ('structure.rst', 'structure.inpcrd', 'structure.mdcrd'),
                    'parmfile': ('structure.prmtop',)}

        for field in defaults:
            userval = self.amber.__dict__[field]
            if userval is not None and not os.path.exists(os.path.join(inputdir, userval)):
                self.amber.__dict__[field] = None

            if self.amber.__dict__[field] is None:
                for val in defaults[field]:
                    if os.path.exists(os.path.join(inputdir, val)):
                        self.amber.__dict__[field] = val
                        break

            if userval is not None and self.amber.__dict__[field] is not None and self.amber.__dict__[field] != userval:
                logger.warning('Could not locate structure file {}. Using {} instead.'.format(
                    os.path.join(inputdir, userval), os.path.join(
                        inputdir, self.amber.__dict__[field])
                ))
            elif self.amber.__dict__[field] is None:
                raise RuntimeError('Could not locate any {f:} file in {i:}. '
                                   'Please set the Equilibration.acemd.{f:} property to '
                                   'point to the {f:} file'.format(f=field, i=inputdir))

    # FIXME: add inputdir and outputdir to write method
    # need to decide how to organize MD engine input files!!!
    # Probably will create a bash script here with all the information needed to run pmemd, e.g.:
    # pmemd.cuda_SPFP -O -i Equil.in -o Equil.out -c structure.rst -p structure \
    # -r EquilRestart.rst -x EquilRestart.nc (-ref Min.rst)

    def write(self, inputdir, outputdir):
        """ Write the equilibration protocol

        Writes the equilibration protocol and files into a folder for execution
        using files inside the inputdir directory

        Parameters
        ----------
        inputdir : str
            Path to a directory containing the files produced by a build process.
        outputdir : str
            Directory where to write the equilibration setup files.

        Examples
        --------
        >>> md = Equilibration()
        >>> md.write('./build','./equil')
        """
        # this code automatically loops through the amber attributes and assigns the given value to
        # the parameter.

        self._findFiles(inputdir)

        restartfile = os.path.join(inputdir, self.amber.coordinates)
        topology = os.path.join(inputdir, self.amber.parmfile)

        if self.amber.consref is not None:
            reference = os.path.join(inputdir, self.amber.consref)

        # FOR NOW WE WILL HAVE A FIXED STRING LIKE IN THE ACEMD PROTOCOL
        # THIS WAY WE CAN CONTROL BETTER WHAT PARAMETERS CAN BE SPECIFIED
        # IN THE PROTOCOL

        # protocol=[]
        # i=0
        # for key, value in self.amber.__dict__.items():

        #     if not key.startswith('_') and key not in ['bincoordinates', 'coordinates',
        #                                                'consref', 'parmfile','FORTRAN',
        #                                                'bash', 'outputnc']:

        #         # cleans up the .in file a bit
        #         if i % 3 == 0 and i>0:
        #             protocol.append('\n   {}={}'.format(key,value))
        #         else:
        #             protocol.append('{}={}'.format(key,value))

        #         i+=1

        # # format the FORTRAN file with MD parameters
        # self.amber.FORTRAN = self.amber.FORTRAN+'   '+(', '.join(protocol)) + "\n /"

        # preparing the FORTRAN file
        self.amber.FORTRAN = self.amber.FORTRAN.replace(
            'IMIN', str(self.amber.imin))
        self.amber.FORTRAN = re.sub(r'\bNTX\b', str(
            self.amber.ntx), self.amber.FORTRAN)
        self.amber.FORTRAN = self.amber.FORTRAN.replace(
            'IREST', str(self.amber.irest))
        self.amber.FORTRAN = self.amber.FORTRAN.replace(
            'NTC', str(self.amber.ntc))
        self.amber.FORTRAN = self.amber.FORTRAN.replace(
            'NTF', str(self.amber.ntf))
        self.amber.FORTRAN = self.amber.FORTRAN.replace(
            'NSTLIM', str(self.amber.nstlim))
        self.amber.FORTRAN = self.amber.FORTRAN.replace(
            'NTT', str(self.amber.ntt))
        self.amber.FORTRAN = self.amber.FORTRAN.replace(
            'GAMMA_LN', str(self.amber.gamma_ln))
        self.amber.FORTRAN = self.amber.FORTRAN.replace(
            'NTR', str(self.amber.ntr))
        self.amber.FORTRAN = self.amber.FORTRAN.replace(
            'IG', str(self.amber.ig))
        self.amber.FORTRAN = self.amber.FORTRAN.replace(
            'TEMP0', str(self.amber.temp0))
        self.amber.FORTRAN = self.amber.FORTRAN.replace(
            'NTPR', str(self.amber.ntpr))
        self.amber.FORTRAN = self.amber.FORTRAN.replace(
            'NTWR', str(self.amber.ntwr))
        self.amber.FORTRAN = self.amber.FORTRAN.replace(
            'NTWX', str(self.amber.ntwx))
        self.amber.FORTRAN = self.amber.FORTRAN.replace(
            'DT', str(self.amber.dt))
        self.amber.FORTRAN = self.amber.FORTRAN.replace(
            'NTB', str(self.amber.ntb))
        self.amber.FORTRAN = self.amber.FORTRAN.replace(
            'NTP', str(self.amber.ntp))
        self.amber.FORTRAN = self.amber.FORTRAN.replace(
            'CUT', str(self.amber.cut))
        self.amber.FORTRAN = self.amber.FORTRAN.replace(
            'IOUTFM', str(self.amber.ioutfm))
        self.amber.FORTRAN = self.amber.FORTRAN.replace(
            'BAROSTAT', str(self.amber.barostat))
        self.amber.FORTRAN = self.amber.FORTRAN.replace(
            'NTXO', str(self.amber.ntxo))

        # write out the FORTRAN file, for now let's call it Production.in
        with open(os.path.join(outputdir, 'Production.in'), 'w') as text_file:
            text_file.write(self.amber.FORTRAN)

        # prepare the bash file
        # putting these strings as placeholders for now
        self.amber.bash = self.amber.bash.replace('INPUT', 'Production.in')
        self.amber.bash = self.amber.bash.replace(
            'OUTPUT', 'Production.out')
        self.amber.bash = self.amber.bash.replace(
            'TOPOLOGY', self.amber.parmfile)
        self.amber.bash = self.amber.bash.replace(
            'RESTART', self.amber.coordinates)
        self.amber.bash = self.amber.bash.replace(
            'TRAJOUT', self.amber.outputnc)
        self.amber.bash = self.amber.bash.replace(
            'RESTOUT', 'Production_new.rst')
        # FIXME: add ref
        # self.amber.bash = self.amber.bash.replace('REFERENCE', self.amber.reference)

        with open(os.path.join(outputdir, 'MD.sh'), 'w') as text_file:
            text_file.write(self.amber.bash)

        shutil.copy(restartfile, outputdir)
        shutil.copy(topology, outputdir)
        # FIXME: add ref
        # shutil.copy(consref, outputdir)



# (c) 2015-2018 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
import numpy as np
from numba import jit

@jit('void(f8[:,:], f8[:,:], f8[:,:], f8[:])', nopython=True, target='cpu')
def _wrapping_dist_numba(X, Y, D, box):
    numX = X.shape[0]
    numY = Y.shape[0]
    numDim = X.shape[1]
    for i in range(numX):
        for j in range(numY):
            d = 0.0
            for k in range(numDim):
                dist = X[i, k] - Y[j, k]
                tmp = dist - box[k] * round(dist / box[k])
                d += tmp * tmp
            D[i, j] = np.sqrt(d)


def _totalContacts(x, *args):
    from moleculekit.util import rotationMatrix
    lipids = args[0]
    headnames = args[1]
    zpos = args[2]
    thresh = args[3]
    neighbours = args[4]
    numlips = args[5]
    box = args[6]
    newxy = np.array(x[:numlips * 2]).reshape((-1, 2))
    rots = x[numlips * 2:]

    allcoords = []
    for i in range(numlips):
        cc = np.squeeze(lipids[i].mol.coords)
        headpos = cc[lipids[i].mol.name == headnames[i]].flatten()[np.newaxis, :]
        cc = cc - headpos  # Center it on the head position
        newloc = np.hstack((newxy[i], zpos[i]))[np.newaxis, :]  # Calculate new head position

        # Doing the rotation
        M = rotationMatrix([0, 0, 1], np.deg2rad(rots[i]))
        newcoords = np.dot(cc, np.transpose(M))

        allcoords.append(newcoords + newloc)

    allcoords = np.array(allcoords, dtype=object)
    numcontacts = 0
    for i in range(len(allcoords)):
        if len(neighbours[i]) == 0:
            continue
        neighcoor = np.vstack(allcoords[neighbours[i]])
        dists = np.zeros((allcoords[i].shape[0], neighcoor.shape[0]))
        _wrapping_dist_numba(allcoords[i], neighcoor, dists, np.array(box, dtype=float))
        numcontacts += np.count_nonzero(dists < thresh)

    return numcontacts


class _RandomDisplacementBounds(object):
    """random displacement with bounds"""

    def __init__(self, bounds, stepsizes):
        self.bounds = bounds
        self.stepsizes = stepsizes

    def __call__(self, x):
        """take a random step but ensure the new position is within the bounds"""
        xnew = np.clip(x + (np.random.uniform(size=x.shape) - 0.5) * self.stepsizes, self.bounds[:, 0],
                       self.bounds[:, 1])
        return xnew


def minimize(lipids, box, stepxy=0.5, steprot=50, contactthresh=2.6):
    from scipy.optimize import basinhopping
    # rotate in 10deg increments
    # translate in a 2x2 box in 0.25A increments
    # swap lipid conformer?
    headnames = [l.headname for l in lipids]
    zpos = [l.xyz[2] for l in lipids]
    neighbours = [l.neighbours for l in lipids]

    pos = np.vstack([l.xyz[:2] for l in lipids])
    x0 = pos.flatten().tolist()
    numlips = len(lipids)
    bounds = [(x - 1, x + 1) for x in x0]
    x0 += [180] * numlips  # Add the rotations
    bounds += [(0, 360)] * numlips  # Add the rotations
    stepsizes = np.ones(numlips * 2) * stepxy
    stepsizes = np.hstack((stepsizes, np.ones(numlips) * steprot))  # Add the rotations

    # define the new step taking routine and pass it to basinhopping
    take_step = _RandomDisplacementBounds(np.vstack(bounds), stepsizes=stepsizes)
    minimizer_kwargs = dict(method="L-BFGS-B", bounds=bounds,
                            args=(lipids, headnames, zpos, contactthresh, neighbours, numlips, box))
    res = basinhopping(_totalContacts, x0, minimizer_kwargs=minimizer_kwargs, disp=True, take_step=take_step, niter=10)
    newpos = res.x[:numlips * 2].reshape((-1, 2))
    newrot = res.x[numlips * 2:]
    return newpos, newrot